# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import numpy as np
from GoldenSectionLineSearchAlgorithm import *

# ------------------------------------------------------------------------
# FINITE DIFFERENCE FOR MULTI-VARIABLE OBJECTIVE FUNCTION
# ------------------------------------------------------------------------
def nablaF(x, F, F0):

    # Extract number of design variables
    numDVs = len(x)

    # Initialize the gradient vector
    gradient = np.zeros(numDVs)

    # Loop over design variables and do finite difference
    for i in range(numDVs):

        # Backup of design variable value
        backup  = x[i]

        # Set the perturbation as a fraction of the x-value
        dx = 0.001
        if x[i] != 0.0:
            dx = dx * x[i]

        # Perturbed vector of design variables
        x[i] += dx

        # Perturbed function value
        perturbedF = F(x)

        # Sought derivative
        gradient[i] = (perturbedF - F0) / dx

        # Reset the design variable
        x[i] = backup

    # Return the gradient vector
    return gradient

# ------------------------------------------------------------------------
# STEEPEST DESCENT SEARCH DIRECTION
# ------------------------------------------------------------------------
def steepestDescentSearchDirection(gradient):

    return np.multiply(gradient, -1.0)

# ------------------------------------------------------------------------
# CONJUGATE GRADIENT SEARCH DIRECTION
# ------------------------------------------------------------------------
def conjugateGradientSearchDirection(gradient, previousGradient):

    # Calculate the suared norm of the gradients
    thisGradientNormSquared = (np.linalg.norm(gradient))**2
    previousGradientNormSquared = (np.linalg.norm(previousGradient))**2

    steepest = np.multiply(gradient, -1.0)
    if previousGradientNormSquared == 0.0:

        return steepest

    else:
        return steepest + np.multiply(previousGradient, -thisGradientNormSquared / previousGradientNormSquared)

# ------------------------------------------------------------------------
# QUASI NEWTON SEARCH DIRECTION
# ------------------------------------------------------------------------
def quasiNewtonSearchDirection(x, oldPoint, gradient, oldGradient, HinvOld, type):

    # Initial declarations
    numVars = len(gradient)

    # If it is the first step, use steepest descent
    if np.linalg.norm(oldGradient) == 0.0:

        Hinv = np.multiply(np.identity(numVars), 0.001)
        searchDirection = np.multiply(gradient, -0.001)

        return [searchDirection, Hinv]

    else:

        deltaX = x - oldPoint
        deltaGrad = gradient - oldGradient

        if type == "DFP":

            deltaXOuterDeltaX = np.outer(deltaX, deltaX)
            deltaXDotDeltaGrad = deltaX.dot(deltaGrad)

            # Hinv = HinvOld + deltaXOuterDeltaX / deltaXDotDeltaGrad
            #      - (HinvOld * deltaGrad *outer* deltaGrad * HinvOld) / (deltaGrad * HinvOld * deltaGrad)
            Hinv = HinvOld + np.divide(deltaXOuterDeltaX , deltaXDotDeltaGrad) \
                   - np.divide(   (np.outer( HinvOld.dot(deltaGrad) , deltaGrad)).dot(HinvOld) , \
                                  (deltaGrad.dot(HinvOld)).dot(deltaGrad)   )

        elif type == "Broyden":

            # Hinv = HinvOld + ((deltaX - HinvOld * deltaGrad) *outer* deltaX * HinvOld)
            #      / (deltaX * HinvOld * deltaGrad)
            Hinv = HinvOld + np.divide(( np.outer( deltaX - (HinvOld.dot(deltaGrad)), deltaX).dot(HinvOld)) , \
                                       (deltaX.dot(HinvOld)).dot(deltaGrad)  )

        elif type == "SR1":

            # Hinv = HinvOld + ((deltaX - HinvOld * deltaGrad) *outer* (deltaX - HinvOld * deltaGrad))
            #      / ((deltaX - HinvOld * deltaGrad) * deltaGrad)
            Hinv = HinvOld + np.divide( np.outer( (deltaX - HinvOld.dot(deltaGrad)) , \
                 ((deltaX - HinvOld.dot(deltaGrad)))) , (deltaX - (HinvOld.dot(deltaGrad))).dot(deltaGrad) )

        elif type == "BFGS":

            Eye = np.identity(numVars)
            deltaXOuterDeltaGrad = np.outer(deltaX, deltaGrad)
            deltaXDotDeltaGrad = deltaX.dot(deltaGrad)
            deltaXOuterDeltaX = np.outer(deltaX, deltaX)

            # Hinv = (Eye - deltaXOuterDeltaGrad / deltaXDotDeltaGrad) * HinvOld * (Eye - deltaXOuterDeltaGrad
            #      / deltaXDotDeltaGrad) + deltaXOuterDeltaX / deltaXDotDeltaGrad
            Hinv = (Eye - (np.divide(deltaXOuterDeltaGrad , deltaXDotDeltaGrad)).dot(HinvOld)).dot(Eye \
                 - np.divide(deltaXOuterDeltaGrad , deltaXDotDeltaGrad)) + deltaXOuterDeltaX / deltaXDotDeltaGrad

        else:
            print("Wrong type given to quasi-Newton search direction algorithm")

    # Calculate the search direction -Hinv*grad
    searchDirection = np.multiply(-1, Hinv.dot(gradient))

    return [searchDirection, Hinv]

# ------------------------------------------------------------------------
# INPUT
# ------------------------------------------------------------------------
def F(x):

    x1 = x[0]
    x2 = x[1]

    k1 = 100.0
    k2 = 90.0
    Fx1 = 20.0
    Fx2 = 40.0

    return k1 * ( np.sqrt( x1**2 + (x2+1)**2 ) - 1 )**2 + k2 * ( np.sqrt( x1**2 + (x2-1)**2 ) - 1 )**2 - ( Fx1 * x1 + Fx2 * x2 )

# Start point
startPoint = np.zeros(2)

# ------------------------------------------------------------------------
# OPTIMIZATION ALGORITHM
# ------------------------------------------------------------------------

# Algorithm parameters
tolerance = 0.05
maxSteps = 100

# Initial declarations
maxFunctionValue = 0.0
minFunctionValue = 0.0
maxDesignVariableValue = 0.0
minDesignVariableValue = 0.0
numDVs = len(startPoint)
searchDirection = np.zeros(numDVs)
previousPoint = np.zeros(numDVs)
previousGradient = np.zeros(numDVs)
inverseHessian = np.zeros((numDVs, numDVs))
previousInverseHessian = np.zeros((numDVs, numDVs))
x = startPoint.copy()

# Start the loop
loopCounter = 0
while loopCounter < maxSteps:

    loopCounter += 1

    # Evaluate objective function and its gradient
    Fvalue = F(x)
    Fgradient = nablaF(x, F, Fvalue)
    gradientNorm = np.linalg.norm(Fgradient)

    # Output
    print('\n'"At step %d of the optimization analysis:" %loopCounter)
    print("Objective function:   ", Fvalue)
    print("Design variables:     ", x)
    print("Gradient vector:      ", Fgradient)
    print("Convergence criterion:", gradientNorm)

    # Check convergence
    if gradientNorm < tolerance:

        print('\n'"The opimization analysis converged")
        break

    elif loopCounter==maxSteps:

        print('\n'"Maximum number of steps reached before convergence")

    else:

        # Determine search direction by steepest descent
        #searchDirection = steepestDescentSearchDirection(Fgradient)

        # Determine search direction by the conjugate gradient method
        #searchDirection = conjugateGradientSearchDirection(Fgradient, previousGradient)

        # Determine the search direction by the quasi Newton method (BFGS, DFP, Broyden, SR1)
        [searchDirection, inverseHessian] = quasiNewtonSearchDirection(x, previousPoint, Fgradient, previousGradient, previousInverseHessian, "BFGS")

        # Store the previous gradient and perhaps Hessian, because some search direction algorithms need it
        previousPoint = x.copy()
        previousGradient = Fgradient.copy()
        previousInverseHessian = inverseHessian.copy()

        # Define the merit function for the line search along the search direction
        def meritFunction(stepSize):

            # Determine the trial point corresponding to the step size
            xTrial = x + stepSize * searchDirection

            # Evaluate the objective function
            FTrial = F(xTrial)

            # Evaluate the merit function
            return FTrial

        # Search along the search direction for the optimal step size
        stepSize = goldenSectionLineSearch(meritFunction, 0.0, 10.0, 50.0, 0.0001, 0.001)

        # Take the step
        x += stepSize * searchDirection