# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Notation:
# ndof     = number of free DOFs
# ntot     = total number of all  DOFs
# free     = range of free DOFs (1 to ndof)
# nelem    = number of elements in elemlist
# elemlist = list of elements
# ua       = vector of displacements for all DOFs
# ug       = vector of displacements for global element DOFs
# Fa       = nodal load vector in model for all DOFs
# Fa_tilde = restoring force vector for all DOFs
# Ka       = stiffness matrix for all DOFs
# Kf       = stiffness matrix for free DOFs
# R        = residual force vector (Fa_tilde-Fa) for free DOFs

import numpy as np

def linearStaticAnalysis(model, trackNode, trackDOF, DDMparameters, plotFlag=False):

    # Plot the model
    if plotFlag:
        model.plotModel()

    # Get data from the model
    ndof, ntot, Fa, M, elemlist = model.getData()
    free = range(ndof)
    nelem = len(elemlist)

    # Initialize global vectors and matrix
    Fa_tilde = np.zeros(ntot)
    ua = np.zeros((ntot,3))
    Ka = np.zeros((ntot, ntot))

    # Initialize element state
    for i in range(nelem):
        id, xyz, ug = model.localize(i, ua)
        element = elemlist[i]
        element.initialize(xyz)

    # Loop over elements to assemble stiffness matrix and load vector
    for i in range(nelem):
        id, xyz, ug = model.localize(i, ua)
        element = elemlist[i]
        Fg_tilde, Kg = element.state(xyz, ug, 1.0)
        Ka[np.ix_(id, id)] = Ka[np.ix_(id,id)] + Kg
        Fa_tilde[id] = Fa_tilde[id] + Fg_tilde

    # Stiffness matrix in the Final configuration
    Kf = Ka[np.ix_(free, free)]
    det = np.linalg.det(Kf)
    if det < 1.0e-10:
        print('\n'"ERROR: The determinant of the stiffness matrix is", det)
        import sys
        sys.exit()

    # Residual vector (Ku-F) in the Final configuration
    Rf = Fa_tilde[free] - Fa[free]

    # Solve the system of equations (linear analysis, so use only first column of ua)
    ua[free, 0] = np.linalg.solve(Kf, -Rf)

    # Set and commit the state to get responses
    for i in range(nelem):
        id, xyz, ug = model.localize(i, ua)
        element = elemlist[i]
        element.state(xyz, ug, 1.0)
        element.commit(xyz, ug)

    # DDM sensitivity analysis using the "adjoint method"
    sensitivity = []
    dTrackedDua = np.zeros(ntot)
    index = model.DOF[trackNode - 1, trackDOF - 1]
    dTrackedDua[index] = 1
    adjointVector = np.linalg.solve(Kf, dTrackedDua[free])
    displacement = np.dot(adjointVector, -Rf)
    if len(DDMparameters) > 0:
        for ddmIndex in range(len(DDMparameters)):
            ddmRHSa = np.zeros(ntot)
            if DDMparameters[ddmIndex][0] == 'Element':
                for eleNum in DDMparameters[ddmIndex][2]:
                    i = eleNum-1
                    id, xyz, ug = model.localize(i, ua)
                    element = elemlist[i]
                    ddmRHSg, ddmKg, dKgdug = element.stateDerivative(xyz, ug, 1.0, DDMparameters[ddmIndex][1], 0, True)
                    # Equivalent: ddmRHSa[id]=ddmRHSa[id]-ddmKg.dot(ug) but then missing q-contribution to dF/dtheta
                    ddmRHSa[id] = ddmRHSa[id] - ddmRHSg
            elif DDMparameters[ddmIndex][0] == 'Nodal load':
                loadIndex = model.DOF[DDMparameters[ddmIndex][1] - 1, DDMparameters[ddmIndex][2] - 1]
                ddmRHSa[loadIndex] = 1
            else:
                print("Cannot handle DDM parameter of type", DDMparameters[ddmIndex][0])
                import sys
                sys.exit()
            sensitivity.append(np.dot(adjointVector, ddmRHSa[free]))

    # Print displacements and element response
    print("--------------------------------------------------------")
    print("Linear analysis completed")
    model.printDOF(ua[:, 0])
    model.printElemResp(ua[:, 0])

    # Print response sensitivity results
    if len(DDMparameters) > 0:
        print("--------------------------------------------------------")
        print("Response sensitivity results:")
        print("--------------------------------------------------------")
        print("          Tracked response:  %.5e" % displacement)
        for ddmIndex in range(len(DDMparameters)):
            print("d(response)/d(parameter %i):  %.5e" % (ddmIndex+1, sensitivity[ddmIndex]))

    # Plot the deformed model
    if plotFlag:
        model.plotDeformedModel(ua[:, 0])

    return displacement, sensitivity

