# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Notation:
# nsteps       = number of load steps
# maxiter      = max. equilibrium iterations per load step
# KcalcFrequency = 0=initial stress method, 1=Newton-Raphson, maxIter=Modified NR
# tol          = tolerance for equilibrium iterations
# ndof         = number of DOFs
# ntot         = total number of all  DOFs
# free         = range of free DOFs (1:ndof)
# nelem        = number of elements in elemlist
# elemlist     = list of elements
# increment    = current load step, i.e., increment
# dlamb        = load increment
# theLambda    = load factor for current load step
# iter         = counter of equilibrium iterations
# ua           = matrix of displacements for "all" DOFs
#                First column of ua  = total displacements
#                Second column of ua = displacements since last committed state
#                Third column of ua  = solution to system of equations
# ua1old       = previously committed displacements (used only if solution does not converge)
# Fref         = nodal load vector in model for all DOFs
# Fa           = scaled nodal load vector for all DOFs (lambda*Fref)
# Fa_tilde     = restoring force vector for all DOFs
# Rf           = residual force vector for free DOFs
# residualNorm = 2-norm of R (Euclidean norm)
# Ka           = tangent stiffness matrix for all DOFs
# Kf           = tangent stiffness matrix for free DOFs

import numpy as np
import math
from scipy.linalg import lu_factor, lu_solve
import matplotlib.pyplot as plt

# ------------------------------------------------------------------------
# TIME SERIES
# ------------------------------------------------------------------------
def timeSeries(t):
    if t<1.0:
        theLambda = t
    elif t<3.0:
        theLambda = 1.0-(t-1.0)
    else:
        theLambda = -1.0+(t-3.0)
    return theLambda

# ------------------------------------------------------------------------
# NONLINEAR STATIC ANALYSIS ALGORITHM
# ------------------------------------------------------------------------
def nonlinearStaticAnalysis(model, nsteps, dt, maxiter, KcalcFrequency, tol, trackNode, trackDOF, DDMparameters, plotFlag=True):

    # Plot the model
    if plotFlag:
        model.plotModel()

    # Get data from the model
    ndof, ntot, Fref, M, elemlist = model.getData()
    free = range(ndof)
    nelem = len(elemlist)

    # Initialize restoring force vector, stiffness matrix, and displacement vectors
    Fa_tilde = np.zeros(ntot)
    Ka = np.zeros((ntot, ntot))
    ua = np.zeros((ntot, 3))
    ua1old = np.zeros(ntot)
    if len(DDMparameters) > 0:
        ddmua = np.zeros(ntot)

    # Index of tracked DOF
    index = model.DOF[trackNode - 1, trackDOF - 1]

    # Initialize element state
    for i in range(nelem):
        id, xyz, ug = model.localize(i, ua[:,0])
        elemlist[i].initialize(xyz)

    # Create the plot
    plotDisp = []
    plotLambda = []
    plotDisp.append(0.0)
    plotLambda.append(0.0)
    if plotFlag:
        plt.ion()
        plt.figure()
        plt.title('Converged Newton-Raphson Steps for Tracked DOF')
        plt.grid(True)
        plt.autoscale(True)
        plt.ylabel('Load Factor')
        plt.xlabel('Displacement')

    # Print header for summary of load steps
    print("--------------------------------------------------------")
    print("Step     Lambda   Iterations   Residual    DOF tracking")
    print("--------------------------------------------------------")

    # Loop over load steps
    time = 0.0
    trackAllDDMs = np.zeros((nsteps, len(DDMparameters)))
    for increment in range(nsteps):

        # Increment time
        time += dt

        # Reset the Newton-Raphson iteration counter for every increment
        iteration = 0

        # Determine load factor from time series
        theLambda = timeSeries(time)

        # Multiply the external loads by the load factor
        Fa = theLambda * Fref

        # Perform equilibrium iterations for load step
        m = KcalcFrequency
        for i in range(maxiter):

            # Loop over elements to assemble tangent stiffness and restoring force
            Fa_tilde[:] = 0.0
            if m == KcalcFrequency and not (m == 0 and increment > 0):
                Ka[:] = 0.0
            for j in range(nelem):
                id, xyz, ug = model.localize(j, ua)
                element = elemlist[j]
                Fg_tilde, Kg = element.state(xyz, ug, theLambda)
                Fa_tilde[id] = Fa_tilde[id] + Fg_tilde

                # Check if user wants Modified Newton-Raphson
                if m == KcalcFrequency and not (m == 0 and increment > 0):
                    Ka[np.ix_(id, id)] = Ka[np.ix_(id, id)] + Kg

            # Reset the stiffness calculation flag
            if m == KcalcFrequency and not (m == 0 and increment > 0):
                m = 0

            # Compute residual
            Rf = Fa_tilde[free] - Fa[free]

            # Exit equilibrium loop if 2-norm of residual vector is less than tolerance
            residualNorm = np.linalg.norm(Rf)
            if residualNorm < tol:
                break

            # Solve system of equations
            Kf = Ka[np.ix_(free, free)]
            det = np.linalg.det(Kf)

            # Check for zero determinant
            if np.abs(det) < 1.0e-10:
                print('\n'"Error: The determinant of the stiffness matrix is", det)
                return plotLambda[1:len(plotLambda)], plotDisp[1:len(plotDisp)], trackAllDDMs

            # Warning for negative determinant
            if det < 1.0e-10:
                print('\n'"Warning: The determinant of the stiffness matrix is", det)

            ua[free, 2] = np.linalg.solve(Kf, -Rf)

            # Update displacements
            ua[:, 0] = ua[:, 0] + ua[:, 2]
            ua[:, 1] = ua[:, 1] + ua[:, 2]

            m += 1
            iteration += 1

        # Check that equilibrium iteration loop converged. If not, break load
        # step loop after reverting to previous load step that did converge.
        if residualNorm > tol:
            print('\n'"No convergence with residual", residualNorm, ">", tol, "in", maxiter, "iterations for increment", increment+1)
            #import sys
            #sys.exit()

        # Here we have converged at this increment, so do DDM response sensitivity analysis, before commit
        if len(DDMparameters) > 0:

            # Make sure we have the current tangent stiffness
            Ka[:] = 0.0
            for j in range(nelem):
                id, xyz, ug = model.localize(j, ua)
                element = elemlist[j]
                Fg_tilde, Kg = element.state(xyz, ug, theLambda)
                Ka[np.ix_(id, id)] = Ka[np.ix_(id, id)] + Kg

            # Do LU factorization
            Kf = Ka[np.ix_(free, free)]
            LU = lu_factor(Kf)

            for ddmIndex in range(len(DDMparameters)):

                # Reset the right-hand side for each parameter
                ddmRHSa = np.zeros(ntot)

                # Gather right-hand side contributions from elements
                if DDMparameters[ddmIndex][0] == 'Element':

                    # Check which elements have the parameter
                    ddmIsHere = np.full(nelem, False)
                    for eleNum in DDMparameters[ddmIndex][2]:
                        ddmIsHere[eleNum - 1] = True

                    # Must call all elements, even if the parameter isn't in them
                    for i in range(nelem):
                        id, xyz, ug = model.localize(i, ua)
                        element = elemlist[i]
                        dFg, dKg, dKgdug = element.stateDerivative(xyz, ug, theLambda, DDMparameters[ddmIndex][1], ddmIndex, ddmIsHere[i])
                        ddmRHSa[id] = ddmRHSa[id] - dFg

                elif DDMparameters[ddmIndex][0] == 'Nodal load':
                    loadIndex = model.DOF[DDMparameters[ddmIndex][1] - 1, DDMparameters[ddmIndex][2] - 1]
                    ddmRHSa[loadIndex] = 1
                    for i in range(nelem):
                        id, xyz, ug = model.localize(i, ua)
                        element = elemlist[i]
                        dFg, dKg, dKgdug = element.stateDerivative(xyz, ug, theLambda, DDMparameters[ddmIndex][1], ddmIndex, False)
                        ddmRHSa[id] = ddmRHSa[id] - dFg

                else:
                    print("Cannot handle DDM parameter of type", DDMparameters[ddmIndex][0])
                    import sys
                    sys.exit()

                # Solve for the derivative of the displacements
                ddmRHSf = ddmRHSa[free]
                ddmDisplacementNew = lu_solve(LU, ddmRHSf)

                # Pick up the derivative that will be plotted
                ddmua[free] = ddmDisplacementNew
                trackAllDDMs[increment, ddmIndex] = ddmua[index]

                # Calculate unconditional sensitivities
                # (Remember, each material must store history variables for ALL ddm variables)
                if DDMparameters[ddmIndex][0] == 'Element':
                    for i in range(nelem):
                        id, xyz, ug = model.localize(i, ua)
                        id, xyz, ddmug = model.localize(i, ddmua)
                        element = elemlist[i]
                        element.commitSensitivity(xyz, ug, ddmug, DDMparameters[ddmIndex][1], ddmIndex, ddmIsHere[i])
                elif DDMparameters[ddmIndex][0] == 'Nodal load':
                    for i in range(nelem):
                        id, xyz, ug = model.localize(i, ua)
                        id, xyz, ddmug = model.localize(i, ddmua)
                        element = elemlist[i]
                        element.commitSensitivity(xyz, ug, ddmug, DDMparameters[ddmIndex][1], ddmIndex, False)

        # Commit and update element states using converged displacements
        for i in range(nelem):
            id, xyz, ug = model.localize(i, ua)
            element = elemlist[i]
            element.commit(xyz, ug)

        # Print status of load step
        trackDisp = ua[index, 0]
        print("%3d %10.2f %9d %15.3E %12.5f" % (increment + 1, theLambda, iteration, residualNorm, trackDisp))

        # Zero second column of displacement vector after commit
        ua[:, 1] = 0.0

        # Add point to the plot
        plotDisp.append(trackDisp)
        plotLambda.append(theLambda)
        if plotFlag:
            plt.plot(plotDisp, plotLambda, 'ko-', linewidth=1.0)
            plt.pause(0.01)

    # Here we are done with all increments, so print nodal displacements and element forces
    if residualNorm < tol:
        model.printDOF(ua[:, 0])
        model.printElemResp(ua[:, 0])

    # Ask the user to close the plot
    print('\n'"Click somewhere in the plot to continue...")
    if plotFlag:
        plt.waitforbuttonpress()

    # Plot the deformed model
    if plotFlag:
        model.plotDeformedModel(ua[:, 0])

    # Remove the first zero-entry when returning displacements because the first point is plotted above
    return plotLambda[1:len(plotLambda)], plotDisp[1:len(plotDisp)], trackAllDDMs
