# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 1: Linear elastic truss element

import numpy as np

class element1():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, E, A, N0, elno):

        self.type = "element1"

        self.no = elno
        self.E = E
        self.A = E
        self.N0 = N0

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ugMatrix, theLambda):

        # Total displacements
        ug = ugMatrix[:, 0]

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        # Kinematic matrix
        Tbg = np.array([-dx[0], -dx[1], dx[0], dx[1]])

        # Basic stiffness coefficient
        Kb = self.E * self.A / L

        # Deformation and force in the basic configuration
        ub = Tbg.dot(ug)
        Fb = Kb * ub + self.N0

        # Store basic force to avoid recomputing in getResponse
        self.FbTrial = Fb

        # Restoring force and stiffness matrix in the global configuration
        Fg_tilde = Tbg.dot(Fb)
        Kg = np.outer(Tbg.dot(Kb), Tbg)

        return Fg_tilde, Kg

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, ug):

        # Commit basic forces for getResponse
        self.FbCommitted = self.FbTrial

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Print element force
        N = self.FbCommitted
        print("Element %d (%s): N=%.1f " % (self.no, self.type, N))
