# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 13: Force-based beam-column element

import numpy as np
from G2Quadrature import *
from G2SectionWideFlange import *
from G2SectionRectangular import *

class element13():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, nsec, q, section, material, elno):

        self.type = "element13"
        self.q = q
        self.nsec = nsec

        # Create sections
        sections = []
        for i in range(self.nsec):
            if section[0] == 'WideFlange':
                hw = section[1]
                bf = section[2]
                tf = section[3]
                tw = section[4]
                nf = section[5]
                nw = section[6]
                sec = wfsection(hw, bf, tf, tw, nf, nw, material)
                sections.append(sec)
            elif section[0] == 'Rectangular':
                h = section[1]
                b = section[2]
                nfib = section[3]
                sec = rectangularSection(h, b, nfib, material)
                sections.append(sec)

        # Get integration parameters for an integration rule
        self.xip, self.weight = Gauss(self.nsec)

        # Set other instantiation variables
        self.no = elno
        self.secs = sections

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = np.zeros(3)

        self.barFbTrial = []
        self.barFbCommitted = np.zeros(3)


    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'q':
            self.q = value
        else:
            for i in range(self.nsec):
                self.secs[i].setParameter(parameter, value)


    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, U, theLambda):

        # It is a bit awkward to pass all columns of U through the compatibility and
        # equilibrium iterations; for that reason, work with *incremental* deformations
        ug = U[:, 1]

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        # Form transformation matrix
        Tbg = np.array([[-dx[1]/L, dx[0]/L, 1.0, dx[1]/L, -dx[0]/L, 0.0],
                      [-dx[1]/L, dx[0]/L, 0.0, dx[1]/L, -dx[0]/L, 1.0],
                      [-dx[0],  -dx[1],   0.0, dx[0],    dx[1],   0.0]])

        # Incremental basic deformations
        ub = Tbg.dot(ug)

        # Compute the Basic forces that produce section deformations compatible with ub
        Fb, Kb = self.compatibility(ub, L, theLambda)

        # Store trial basic forces
        self.FbTrial = Fb

        # Global restoring force and stiffness matrix
        q = theLambda * self.q
        shear = q*L/2.0
        Fg = (np.transpose(Tbg)).dot(Fb) + np.array([-dx[1]*shear, dx[0]*shear, 0.0, -dx[1]*shear, dx[0]*shear, 0.0])
        Kg = (np.transpose(Tbg).dot(Kb)).dot(Tbg)

        return Fg, Kg


    # -------------------------------------------------
    # Compatibility iterations
    # -------------------------------------------------
    def compatibility(self, ub, L, theLambda):

        # Set iteration controls
        maxiter = 100
        tol = 1e-5

        # Integration points and weights
        points = self.xip
        weights = self.weight

        # Interpolation functions for equilibrium evaluated at integration points
        x = []
        tildeTsb1 = []
        tildeTsb2 = []
        tildeTsb3 = []
        for i in range(self.nsec):
            x.append( (points[i] + 1.0) / 2.0 )
            tildeTsb1.append( -(1.0 - x[i]) )
            tildeTsb2.append( x[i] )
            tildeTsb3.append( 1.0 )

        # Initialize vector of incremental forces for iteration
        FbIncrement = np.zeros(3)

        # Element load
        q = theLambda * self.q
        mom = q * L**2 / 12.0
        barFb = [mom, -mom, 0.0]
        self.barFbTrial = barFb

        # Iterate on section forces to determine compatibility
        for j in range(maxiter):

            # Residual deformations and flexibility by summing over integration points
            ubResidual = np.copy(ub)
            fb = np.zeros((3, 3))
            for i in range(self.nsec):

                # Get the section
                section = self.secs[i]

                # Equilibrium transformation between Section and Basic
                tildeTsb = np.array([[tildeTsb1[i], tildeTsb2[i], 0.0],
                                     [0.0, 0.0, tildeTsb3[i]]])

                # Section forces from basic forces
                parabola = q * x[i] * (x[i] - 1) * L * L / 2
                Fs = tildeTsb.dot(FbIncrement + self.FbCommitted + np.subtract(barFb, self.barFbCommitted)) - np.array([parabola, 0.0])

                # Section state determination
                us, Ks = section.equilibrium(Fs)

                # Section contribution to residual deformations and flexibility
                ubResidual = ubResidual - np.transpose(tildeTsb).dot(us) * (weights[i] * L/2)
                fb = fb + (np.transpose(tildeTsb).dot(np.linalg.inv(Ks))).dot(tildeTsb) * (weights[i]*L/2)

            FbIncrement = np.linalg.solve(fb, ubResidual) + FbIncrement

            # Make sure everything gets done if linear analysis
            if np.linalg.norm(ubResidual) < tol:
                break

        if np.linalg.norm(ubResidual) >= tol:
            print("ERROR: Compatibility iterations in element", self.no, "failed")
            import sys
            sys.exit()

        Fb = self.FbCommitted + FbIncrement + np.subtract(barFb, self.barFbCommitted)
        Kb = np.linalg.inv(fb)

        return Fb, Kb


    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        # Commit section state with converged us
        for i in range(self.nsec):
            section = self.secs[i]
            section.commit()

        # Commit basic forces for element
        self.FbCommitted = self.FbTrial
        self.barFbCommitted = self.barFbTrial


    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, xyz, ug, theLambda, ddmParameter):

        return 0.0, 0.0


    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, xyz, ug, ddmug, ddmParameter):

        return


    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Geometry
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Print element forces
        shear = (self.FbCommitted[0] + self.FbCommitted[1]) / L
        [N1, N2] = self.FbCommitted[2] * np.array([-1, 1])
        [V1, V2] = shear * np.array([1, - 1])
        [M1, M2] = [self.FbCommitted[0], self.FbCommitted[1]]

        print("Element %d (%s): N1=%.1f / N2=%.1f / V1=%.1f / V2=%.1f / M1=%.1f / M2=%.1f" % (self.no, self.type, N1, N2, V1, V2, M1, M2))