# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 2: Nonlinear truss element with path-dependent material

from G2MaterialBilinear import *
from G2MaterialPlasticity import *
from G2MaterialBoucWen import *

class element2():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, N0, section, material, elno):

        self.type = "element2"

        # Store input
        self.no = elno
        self.N0 = N0
        self.A = float(section[1])

        # Create material
        if material[0] == 'Bilinear':
            self.fiber = bilinearMaterial(material)
        elif material[0] == 'Plasticity':
            self.fiber = plasticityMaterial(material)
        elif material[0] == 'BoucWen':
            self.fiber = boucWenMaterial(material)
        else:
            print("Error: Wrong material given to wide-flange section")
            import sys
            sys.exit()

        # Store basic forces to ease getResponse
        self.FbTrial = []
        self.FbCommitted = []

    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'q':
            self.q = value
        elif parameter == 'A':
            self.A = value
        else:
            self.fiber.setParameter(parameter, value)

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ug, theLambda):

        # Element length
        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))

        # Direction cosines
        dx = dx / L

        # Transformation matrix
        Tbg = np.array([-dx[0], -dx[1], dx[0], dx[1]])

        # Basic deformations
        ub = Tbg.dot(ug)

        # Strain
        epsilon = ub/L

        # Material state determination
        sigma, E = self.fiber.state(epsilon)

        # Basic stiffness
        Kb = E * self.A / L

        # Basic force
        Fb = sigma * self.A + self.N0

        # Store basic forces to avoid recomputing in getResponse
        self.FbTrial = Fb

        # Global restoring force and stiffness matrix
        Fg = np.multiply(Fb, Tbg)
        Kg = np.multiply(Kb, np.outer(Tbg, Tbg))

        return Fg, Kg

    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, xyz, ug, theLambda, ddmParameter, ddmIndex, ddmIsHere, dKflag='none'):

        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L
        Tbg = np.array([-dx[0], -dx[1], dx[0], dx[1]])

        ub = Tbg.dot(ug)
        epsilon = ub/L

        sigma, E = self.fiber.state(epsilon)
        dsigma, dE, dEdeps = self.fiber.stateDerivative(epsilon, ddmParameter, ddmIndex, ddmIsHere, dKflag)

        dArea = 0.0
        if ddmParameter=='A' and ddmIsHere:
            dArea = 1.0

        dKb = dE * self.A / L + E * dArea / L
        dFb = dsigma * self.A + sigma * dArea

        dFg = Tbg.dot(dFb)
        dKg = np.outer(Tbg.dot(dKb), Tbg)

        A, B, C = np.ix_(Tbg, Tbg, Tbg)
        dKgdug = np.multiply(dEdeps * self.A / L**2, A * B * C)

        return dFg, dKg, dKgdug

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        self.FbCommitted = self.FbTrial
        self.fiber.commit()

    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, xyz, ug, ddmug, ddmParameter, ddmIndex, ddmIsHere):

        dx = xyz[1,:] - xyz[0,:]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L
        Tbg = np.array([-dx[0], -dx[1], dx[0], dx[1]])

        ub = Tbg.dot(ug)
        epsilon = ub/L

        ddmub = Tbg.dot(ddmug)
        ddmepsilon = ddmub/L

        self.fiber.commitSensitivity(epsilon, ddmepsilon, ddmParameter, ddmIndex, ddmIsHere)

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        # Print element force
        N = self.FbCommitted
        print("Element %d (%s): N=%.1f " % (self.no, self.type, N))
