# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Element 8: Two-component parallel-model concentrated hinge element

import numpy as np

class element8():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, E, A, I, My, alpha, elno):

        self.type = "element8"

        # Store input
        self.no    = elno
        self.E     = E
        self.A     = A
        self.I     = I
        self.My    = My
        self.alpha = alpha

        # Note: No trial/committed variables are present because the element formulation is
        # in terms of total displacement; path-dependent behavior is not represented

    # -------------------------------------------------
    # Initialize
    # -------------------------------------------------
    def initialize(self, xyz):

        # Geometry restricted to 1,2 plane (x,y)
        dx = xyz[1, :] - xyz[0, :]

        L = np.sqrt(dx.dot(dx))

        # Check length
        if L < 1e-10:
            print('\n'"ERROR: Element number", self.no, "has zero length")
            import sys
            sys.exit()

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, xyz, ugMatrix, theLambda):

        # Extract *TOTAL* displacements (cannot deal with path-dependent behaviour for now)
        ug = np.copy(ugMatrix[:, 0])

        # Geometry
        dx = xyz[1, :] - xyz[0, :]
        L = np.sqrt(dx.dot(dx))
        dx = dx / L

        # Form kinematic matrix
        Tbg = np.array([[-dx[1] / L, dx[0] / L, 1.0, dx[1] / L, -dx[0] / L, 0.0],
                        [-dx[1] / L, dx[0] / L, 0.0, dx[1] / L, -dx[0] / L, 1.0],
                        [-dx[0], -dx[1], 0.0, dx[0], dx[1], 0.0]])

        # Determine basic deformations from global deformations
        ub = Tbg.dot(ug)

        # Compute basic flexural forces and stiffness from a two-component parallel model
        EIL = self.E * self.I / L
        Fbflex, Kbflex = self.twocomponent([ub[0], ub[1]], EIL)

        # Basic stiffness matrix
        EAL = self.E * self.A / L
        Kb = np.array([[Kbflex[0,0], Kbflex[0,1], 0.0],
                       [Kbflex[1,0], Kbflex[1,1], 0.0],
                       [0.0,         0.0,         EAL]])

        # Basic force vector
        Fb = np.zeros(3)
        Fb[0] = Fbflex[0]
        Fb[1] = Fbflex[1]
        Fb[2] = EAL * ub[2]

        # Transform from basic to global
        Fg = (np.transpose(Tbg)).dot(Fb)
        Kg = (np.transpose(Tbg).dot(Kb)).dot(Tbg)

        return Fg, Kg

    # -------------------------------------------------
    # Evaluate two-component parallel model
    # -------------------------------------------------
    def twocomponent(self, ub, EIL):

        # Release codes:
        # release=1 means no hinges
        # release=2 means hinge at left-hand side (End 1)
        # release=3 means hinge at right-hand side (End 2)
        # release=4 means hinge at both ends

        # Force increment if no hinges are present
        firstReleaseCode = 1
        FbIncr, Kb = self.increment(ub, firstReleaseCode, EIL)

        # Scale factor to next hinge formation
        scaleFactor, secondReleaseCode = self.eventFactor([0, 0], FbIncr, firstReleaseCode)

        # Basic forces and deformations in this situation
        Fb = scaleFactor * FbIncr
        ub = np.multiply(ub, (1.0 - scaleFactor))

        # Check if a hinge forms, at either end
        thirdReleaseCode = 0

        if secondReleaseCode == 2 or secondReleaseCode == 3:

            # Force increment if a hinge forms, at either end
            FbIncr, Kb = self.increment(ub, secondReleaseCode, EIL)

            # Scale factor to next hinge formation
            scaleFactor, thirdReleaseCode = self.eventFactor(Fb, FbIncr, secondReleaseCode)

            # Basic forces and deformations in this situation
            Fb = Fb + scaleFactor * FbIncr
            ub = np.multiply(ub, (1.0 - scaleFactor))

        # Check if hinges form at both ends, either from none or one previously
        if secondReleaseCode == 4 or thirdReleaseCode == 4:

            # Force increment if hinges form at both ends
            FbIncr, Kb = self.increment(ub, 4, EIL)

            # Basic forces and deformations in this situation
            Fb = Fb + FbIncr

        return Fb, Kb

    # -------------------------------------------------
    # Increment in basic forces implied by given displacements and hinge situation
    # -------------------------------------------------
    def increment(self, ub, release, EIL):

        KbComponent1 = np.array([[4 * EIL, 2 * EIL],
                                 [2 * EIL, 4 * EIL]])

        if release == 1:

            KbComponent2 = np.array([[4 * EIL, 2 * EIL],
                                     [2 * EIL, 4 * EIL]])

        elif release == 2:

            KbComponent2 = np.array([[0.0, 0.0],
                                     [0.0, 3 * EIL]])

        elif release == 3:

            KbComponent2 = np.array([[3 * EIL, 0.0],
                                     [0.0, 0.0]])

        elif release == 4:

            KbComponent2 = np.zeros((2, 2))

        else:
            print('\n'"Error in the increment method of element number", self.no)
            import sys
            sys.exit()

        # Add contributions from both components
        Kb = self.alpha * KbComponent1 + (1.0 - self.alpha) * KbComponent2

        # Basic forces
        FbIncr = Kb.dot(ub)

        return FbIncr, Kb

    # -------------------------------------------------
    # Scale factor to next hinge formation
    # -------------------------------------------------
    def eventFactor(self, Fb, FbIncr, oldReleaseCode):

        # Determine trial scaling factor at End 1
        if FbIncr[0] != 0:
            trialScaleEnd1 = max((self.My - Fb[0]) / FbIncr[0], (-self.My - Fb[0]) / FbIncr[0])
        else:
            # Flag if the moment here at End 1 is zero
            trialScaleEnd1 = 2

        # Determine trial scaling factor at End 2
        if FbIncr[1] != 0:
            trialScaleEnd2 = max((self.My - Fb[1]) / FbIncr[1], (-self.My - Fb[1]) / FbIncr[1])
        else:
            # Flag if the moment here at End 2 is zero
            trialScaleEnd2 = 3

        # Check where the next hinge will form
        if oldReleaseCode == 1:
            if trialScaleEnd1 < trialScaleEnd2:
                # Hinge at end 1
                newReleaseCode = 2
                scaleFactor = trialScaleEnd1

            elif trialScaleEnd1 > trialScaleEnd2:
                # Hinge at end 2
                newReleaseCode = 3
                scaleFactor = trialScaleEnd2

            elif trialScaleEnd1 == trialScaleEnd2:
                newReleaseCode = 4
                scaleFactor = trialScaleEnd1

            else:
                scaleFactor = 1.0
                newReleaseCode = 1

        elif oldReleaseCode == 2:
            # After old yielding at End 1, the second hinge will form at End 2
            newReleaseCode = 4
            scaleFactor = trialScaleEnd2

        elif oldReleaseCode == 3:
            # After old yielding at End 2, the second hinge will form at End 1
            newReleaseCode = 4
            scaleFactor = trialScaleEnd1

        else:
            print('\n'"Arrived at illegal release option in scale factor method of element", self.no)
            import sys
            sys.exit()

        # Check if there is no changes in hinges, indicated by scale factor greater than unity
        if scaleFactor > 1.0:
            newReleaseCode = oldReleaseCode
            scaleFactor = 1.0

        return scaleFactor, newReleaseCode

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self, xyz, u):

        # No action here because this element has a "total displacement" material model
        return

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self, xyz):

        print("Element %d (%s): No information" % (self.no, self.type))

