# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import numpy as np

class bilinearMaterial():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, mat):

        # Material and geometry properties
        self.E = float(mat[1])      # Modulus of elasticity
        self.fy = float(mat[2])     # Yield stress
        self.alpha = float(mat[3])  # Second slope stiffness is alpha*E

        # Trial instantiation variables
        self.trialStrain = 0.0
        self.trialStress = 0.0
        self.trialBackStress = 0.0
        self.trialYielding = False

        # Committed instantiation variables
        self.committedStrain = 0.0
        self.committedStress = 0.0
        self.committedBackStress = 0.0
        self.committedYielding = False

        # Trial sensitivity history variables
        self.trialStrainDerivative = [0.0]
        self.trialStressDerivative = [0.0]
        self.trialBackStressDerivative = [0.0]

        # Committed sensitivity history variables
        self.committedStrainDerivative = [0.0]
        self.committedStressDerivative = [0.0]
        self.committedBackStressDerivative = [0.0]


    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'E':
            self.E = value
        elif parameter == 'fy':
            self.fy = value
        elif parameter == 'alpha':
            self.alpha = value
        else:
            print("Cannot set parameter", parameter, "in material")

    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, eps):

        # This is an incremental material
        deps = eps[1]

        # Check for unloading
        unloading = False
        if (self.committedYielding == True and (self.committedStress - self.committedBackStress) * deps < 0):
            unloading = True

        # Check if the last state was elastic or if the strain increment implies unloading from yielding
        if self.committedYielding is False or unloading:

            # Strain increment that would cause yielding
            depsToYield = (np.sign(deps) * self.fy + self.committedBackStress - self.committedStress) / self.E

            # Keep elastic state handy, in case that becomes the conclusion
            self.trialStress = self.committedStress + self.E * deps
            Et = self.E
            self.trialYielding = False

            # Check if the strain increment causes yielding from an elastic state (initiation of unloading is elastic)
            if abs(deps) > abs(depsToYield) and not unloading:
                self.trialStress = self.trialStress + (self.alpha * self.E - self.E) * (deps - depsToYield)
                self.trialBackStress = self.committedBackStress + self.alpha * self.E * (deps - depsToYield)
                Et = self.alpha * self.E
                self.trialYielding = True

        else:

            # Continue plastic loading
            self.trialStress = self.committedStress + self.alpha * self.E * deps
            self.trialBackStress = self.committedBackStress + self.alpha * self.E * deps
            Et = self.alpha * self.E
            self.trialYielding = True

        return self.trialStress, Et

    # -------------------------------------------------
    # Equilibrium iterations to find strain for given stress
    # -------------------------------------------------
    def equilibrium(self, stressIncrement):

        # Set iteration controls
        maxiter = 100
        tol = 1e-5

        # Initialize incremental trial strain
        strainIncrement = 0.0

        # Iterate to achieve equilibrium
        for j in range(maxiter):

            self.trialStrain = self.committedStrain + strainIncrement

            trialStressIncrement, stiffness = self.state(np.array([0.0, strainIncrement, 0.0]))
            stressIncrementResidual = stressIncrement - trialStressIncrement
            residualNorm = np.abs(stressIncrementResidual)

            if residualNorm < tol:
                break

            strainIncrement += stressIncrementResidual/stiffness

        if residualNorm >= tol:
            print("ERROR: Equilibrium iterations in bilinear material failed with residual", residualNorm)
            import sys
            sys.exit()

        return strainIncrement, stiffness


    # -------------------------------------------------
    # Initial stiffness
    # -------------------------------------------------
    def initialStiffness(self):

        return self.E

    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, eps, ddmParameter, ddmIndex, ddmIsHere, dKflag='none'):

        # Extend arrays in the first increment
        if ddmIndex > len(self.trialStressDerivative) - 1:
            self.trialStrainDerivative.append(0)
            self.trialStressDerivative.append(0)
            self.trialBackStressDerivative.append(0)
            self.committedStrainDerivative.append(0)
            self.committedStressDerivative.append(0)
            self.committedBackStressDerivative.append(0)

        deps = eps[1]

        unloading = False
        if (self.committedYielding == True and (self.committedStress - self.committedBackStress) * deps < 0):
            unloading = True

        if self.committedYielding is False or unloading:

            depsToYield = (np.sign(deps) * self.fy + self.committedBackStress - self.committedStress) / self.E

            if ddmParameter == 'E' and ddmIsHere:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E - (np.sign(deps) * self.fy + self.committedBackStress - self.committedStress) / self.E**2
            elif ddmParameter == 'fy' and ddmIsHere:
                depsToYieldDerivative = (np.sign(deps) + self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E
            elif ddmParameter == 'alpha' and ddmIsHere:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E
            else:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E

            if ddmParameter == 'E' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] + deps - self.E * self.committedStrainDerivative[ddmIndex]
            elif ddmParameter == 'fy' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] - self.E * self.committedStrainDerivative[ddmIndex]
            elif ddmParameter == 'alpha' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] - self.E * self.committedStrainDerivative[ddmIndex]
            else:
                stressDerivative = self.committedStressDerivative[ddmIndex] - self.E * self.committedStrainDerivative[ddmIndex]

            if ddmParameter == 'E' and ddmIsHere:
                dEt = 1.0
            elif ddmParameter == 'fy' and ddmIsHere:
                dEt = 0.0
            elif ddmParameter == 'alpha' and ddmIsHere:
                dEt = 0.0
            else:
                dEt = 0.0

            if abs(deps) > abs(depsToYield) and not unloading:

                if ddmParameter == 'E' and ddmIsHere:
                    stressDerivative = stressDerivative + (self.alpha - 1) * (deps - depsToYield)  + (self.alpha * self.E - self.E) * (-self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'fy' and ddmIsHere:
                    stressDerivative = stressDerivative  + (self.alpha * self.E - self.E) * (-self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'alpha' and ddmIsHere:
                    stressDerivative = stressDerivative + self.E * (deps - depsToYield)  + (self.alpha * self.E - self.E) * (-self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                else:
                    stressDerivative = stressDerivative  + (self.alpha * self.E - self.E) * (-self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)

                if ddmParameter == 'E' and ddmIsHere:
                    dEt = self.alpha
                elif ddmParameter == 'fy' and ddmIsHere:
                    dEt = 0.0
                elif ddmParameter == 'alpha' and ddmIsHere:
                    dEt = self.E
                else:
                    dEt = 0.0
        else:

            if ddmParameter == 'E' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] + self.alpha * deps+ self.alpha * self.E * (-self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'fy' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] + self.alpha * self.E * (-self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'alpha' and ddmIsHere:
                stressDerivative = self.committedStressDerivative[ddmIndex] + self.E * deps+ self.alpha * self.E * (-self.committedStrainDerivative[ddmIndex])
            else:
                stressDerivative = self.committedStressDerivative[ddmIndex] + self.alpha * self.E * (-self.committedStrainDerivative[ddmIndex])

            if ddmParameter == 'E' and ddmIsHere:
                dEt = self.alpha
            elif ddmParameter == 'fy' and ddmIsHere:
                dEt = 0.0
            elif ddmParameter == 'alpha' and ddmIsHere:
                dEt = self.E
            else:
                dEt = 0.0

        if dKflag == 'Initial':

            if ddmParameter == 'E' and ddmIsHere:
                dEt = 1.0
            elif ddmParameter == 'fy' and ddmIsHere:
                dEt = 0.0
            elif ddmParameter == 'alpha' and ddmIsHere:
                dEt = 0.0
            else:
                dEt = 0.0

        dKdu = 0

        return stressDerivative, dEt, dKdu

    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self):

        self.committedStrain = self.trialStrain
        self.committedStress = self.trialStress
        self.committedBackStress = self.trialBackStress
        self.committedYielding = self.trialYielding

        self.committedStrainDerivative[:] = self.trialStrainDerivative[:]
        self.committedStressDerivative[:] = self.trialStressDerivative[:]
        self.committedBackStressDerivative[:] = self.trialBackStressDerivative[:]

    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, eps, strainSensitivity, ddmParameter, ddmIndex, ddmIsHere):

        deps = eps[1]
        self.trialStrainDerivative[ddmIndex] = strainSensitivity

        unloading = False
        if (self.committedYielding == True and (self.committedStress - self.committedBackStress) * deps < 0):
            unloading = True

        if self.committedYielding is False or unloading:

            depsToYield = (np.sign(deps) * self.fy + self.committedBackStress - self.committedStress) / self.E

            if ddmParameter == 'E' and ddmIsHere:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E - (np.sign(deps) * self.fy + self.committedBackStress - self.committedStress) / self.E ** 2
            elif ddmParameter == 'fy' and ddmIsHere:
                depsToYieldDerivative = (np.sign(deps) + self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E
            elif ddmParameter == 'alpha' and ddmIsHere:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E
            else:
                depsToYieldDerivative = (self.committedBackStressDerivative[ddmIndex] - self.committedStressDerivative[ddmIndex]) / self.E

            if ddmParameter == 'E' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + deps + self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'fy' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'alpha' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            else:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])

            if abs(deps) > abs(depsToYield) and not unloading:

                if ddmParameter == 'E' and ddmIsHere:
                    self.trialStressDerivative[ddmIndex] = self.trialStressDerivative[ddmIndex] + (self.alpha - 1) * (deps - depsToYield) + (self.alpha * self.E - self.E) * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'fy' and ddmIsHere:
                    self.trialStressDerivative[ddmIndex] = self.trialStressDerivative[ddmIndex] + (self.alpha * self.E - self.E) * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'alpha' and ddmIsHere:
                    self.trialStressDerivative[ddmIndex] = self.trialStressDerivative[ddmIndex] + self.E * (deps - depsToYield) + (self.alpha * self.E - self.E) * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                else:
                    self.trialStressDerivative[ddmIndex] = self.trialStressDerivative[ddmIndex]  + (self.alpha * self.E - self.E) * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)

                if ddmParameter == 'E' and ddmIsHere:
                    self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * (deps - depsToYield) + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'fy' and ddmIsHere:
                    self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                elif ddmParameter == 'alpha' and ddmIsHere:
                    self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.E * (deps - depsToYield) + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)
                else:
                    self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex] - depsToYieldDerivative)

        else:

            if ddmParameter == 'E' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.alpha * deps + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'fy' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'alpha' and ddmIsHere:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.E * deps + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            else:
                self.trialStressDerivative[ddmIndex] = self.committedStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])

            if ddmParameter == 'E' and ddmIsHere:
                self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * deps + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'fy' and ddmIsHere:
                self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            elif ddmParameter == 'alpha' and ddmIsHere:
                self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.E * deps + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])
            else:
                self.trialBackStressDerivative[ddmIndex] = self.committedBackStressDerivative[ddmIndex] + self.alpha * self.E * (strainSensitivity - self.committedStrainDerivative[ddmIndex])

        return

    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self):

        return self.committedStress

