# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import matplotlib.pyplot as plt

################## ELEMENT LIBRARY ##########################
from G2Element1 import *   # Element 1: Linear elastic truss element
from G2Element2 import *   # Element 2: Truss element with path-dependent material
from G2Element3 import *   # Element 3: Truss element with moderate displacements
# (to be implemented)      # Element 4: Truss element with geometric AND material nonlinearity
from G2Element5 import *   # Element 5: Linear elastic frame element
from G2Element6 import *   # Element 6: Linear elastic frame element with P-delta
from G2Element7 import *   # Element 7: Concentrated plasticity frame element: Elastic-perfectly-plastic, i.e., end releases
from G2Element8 import *   # Element 8: Concentrated plasticity frame element: Two-component parallel system
from G2Element9 import *   # Element 9: Concentrated plasticity frame element: One-component series system (uniaxial materials)
from G2Element10 import *  # Element 10: Concentrated plasticity frame element: Rigid interior with path-dependent hinge materials
# (to be implemented)      # Element 11: Concentrated plasticity frame element: Finite plastic hinge lengths
from G2Element12 import *  # Element 12: Distributed plasticity frame element: Displacement-based
from G2Element13 import *  # Element 13: Distributed plasticity frame element: Force-based
#############################################################

class model():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, a):

        # Check the input
        self.checkmodel(a)

        # Read input
        self.NODES = np.array(a[0])
        self.CONSTRAINTS = np.array(a[1])
        self.ELEMENTS = a[2]
        self.SECTIONS = a[3]
        self.MATERIALS = a[4]
        self.NODELOAD = np.array(a[5])
        self.NODEMASS = np.array(a[6])

        # Number DOFs and create ID array
        self.DOF, self.nfree, self.ntot = self.dof_numberer()
        self.ID = self.id_numberer()

        # Create elements
        self.ELEMLIST = self.createElements()

    # -------------------------------------------------
    # Check the input
    # -------------------------------------------------
    def checkmodel(self, a):

        numpar = 7

        if len(a) != numpar:
            print('\n'"ERROR:", str(numpar), "entries required in list for creating model")
            import sys
            sys.exit()

        nnodes = np.size(a[0], 0)
        nconstraints = np.size(a[1], 0)
        nel = len(a[2])
        nsec = len(a[3])
        nmat = len(a[4])
        nloads = np.size(a[5], 0)
        dofload = np.size(a[5], 1)
        dofnod = np.size(a[1], 1)

        if nnodes != nconstraints or (nloads > 0 and nloads != nnodes):
            print('\n'"ERROR: Inconsistent number of nodes, loads, and BCs in model")
            import sys
            sys.exit()

        if nel != nmat:
            print('\n'"ERROR: Inconsistent number of elements and materials in model")
            import sys
            sys.exit()

        if nel != nsec:
            print('\n'"ERROR: Inconsistent number of elements and sections in model")
            import sys
            sys.exit()

        if dofload != dofnod:
            print('\n'"ERROR: Inconsistent number of DOFs in Loads and BCs")
            import sys
            sys.exit()

    # -------------------------------------------------
    # Create matrix that numbers the DOFs
    # -------------------------------------------------
    def dof_numberer(self):

        nnodes = np.size(self.CONSTRAINTS, 0)
        ndof = np.size(self.CONSTRAINTS, 1)
        n = nnodes * ndof

        DOF = np.zeros((np.size(self.CONSTRAINTS, 0), np.size(self.CONSTRAINTS, 1)), dtype = int)

        k = 0
        for i in range(nnodes):
            for j in range(ndof):
                if self.CONSTRAINTS[i,j] == 0:
                    DOF[i,j] = k
                    k = k + 1
                else:
                    n = n - 1
                    DOF[i,j] = n

        nfree = k
        ntot = n

        return DOF, nfree, ntot

    # -------------------------------------------------
    # Create matrix that links DOFs and Elements
    # -------------------------------------------------
    def id_numberer(self):

        nelem = len(self.ELEMENTS)
        numNodesPerElement = 2
        dofnod = np.size(self.DOF, 1)
        ID = np.zeros((nelem, 1+numNodesPerElement*dofnod), dtype = int)

        for i in range(nelem):

            num = 0
            lid = []
            for j in range(numNodesPerElement):
                entry = len(self.ELEMENTS[i][:]) - 2 + j
                nod = int(self.ELEMENTS[i][entry])
                if nod > 0:
                    lid.extend(self.DOF[nod-1,:])
                    num = num + dofnod

            ID[i, 0] = num
            ID[i, 1:len(lid)+1] = lid

        return ID

    # -------------------------------------------------
    # Get element ID, coordinates, and displacements
    # -------------------------------------------------
    def localize(self, elem, U):

        n = self.ID[elem, 0]
        id = self.ID[elem, 1:n+1]
        con = self.ELEMENTS[elem][:]
        xyz = []

        numNodesPerElement = 2
        for i in range(numNodesPerElement):
            nod = int(con[len(con)-2+i])
            if nod > 0:
                xyz.append(self.NODES[nod-1,:])

        xyz = np.array(xyz)

        if len(U) == 0:
            ug = 0
        else:
            ug = U[id]

        return id, xyz, ug

    # -------------------------------------------------
    # Create elements
    # -------------------------------------------------
    def createElements(self):

        nelem = len(self.ELEMENTS)
        ellist = []

        for i in range(nelem):

            eltyp = self.ELEMENTS[i][0]
            if eltyp == 1:
                E  = self.ELEMENTS[i][1]
                A  = self.ELEMENTS[i][2]
                N0 = self.ELEMENTS[i][3]
                el = element1(E, A, N0, i+1)
            elif eltyp == 2:
                N0 = self.ELEMENTS[i][2]
                el = element2(N0, self.SECTIONS[i], self.MATERIALS[i], i+1)
            elif eltyp == 3:
                E  = self.ELEMENTS[i][1]
                A  = self.ELEMENTS[i][2]
                N0 = self.ELEMENTS[i][3]
                el = element3(E, A, N0, i+1)
            elif eltyp == 5:
                E = self.ELEMENTS[i][1]
                A = self.ELEMENTS[i][2]
                I = self.ELEMENTS[i][3]
                q = self.ELEMENTS[i][4]
                el = element5(E, A, I, q, i+1)
            elif eltyp == 6:
                option = self.ELEMENTS[i][1]
                E = self.ELEMENTS[i][2]
                A = self.ELEMENTS[i][3]
                I = self.ELEMENTS[i][4]
                q = self.ELEMENTS[i][5]
                el = element6(option, E, A, I, q, i+1)
            elif eltyp == 7:
                E = self.ELEMENTS[i][1]
                A = self.ELEMENTS[i][2]
                I = self.ELEMENTS[i][3]
                My = self.ELEMENTS[i][4]
                el = element7(E, A, I, My, i+1)
            elif eltyp == 8:
                E = self.ELEMENTS[i][1]
                A = self.ELEMENTS[i][2]
                I = self.ELEMENTS[i][3]
                My = self.ELEMENTS[i][4]
                alpha = self.ELEMENTS[i][5]
                el = element8(E, A, I, My, alpha, i+1)
            elif eltyp == 9:
                el = element9(self.SECTIONS[i], self.MATERIALS[i], i+1)
            elif eltyp == 10:
                el = element10(self.SECTIONS[i], self.MATERIALS[i], i+1)
            elif eltyp == 12:
                nsec = self.ELEMENTS[i][1]
                q = self.ELEMENTS[i][2]
                el = element12(nsec, q, self.SECTIONS[i], self.MATERIALS[i], i+1)
            elif eltyp == 13:
                nsec = self.ELEMENTS[i][1]
                q = self.ELEMENTS[i][2]
                el = element13(nsec, q, self.SECTIONS[i], self.MATERIALS[i], i+1)
            else:
                print('\n'"ERROR: Element type", eltyp, "is not yet implemented")
                import sys
                sys.exit()

            ellist.append(el)

        return ellist

    # -------------------------------------------------
    # Set selected parameter values
    # -------------------------------------------------
    def setParameters(self, parameters):

        for m in range(len(parameters)):

            if parameters[m][0] == 'Element':
                for eleNum in parameters[m][2]:
                    i = eleNum - 1
                    element = self.ELEMLIST[i]
                    element.setParameter(parameters[m][1], parameters[m][3])


    # -------------------------------------------------
    # Print element responses
    # -------------------------------------------------
    def printElemResp(self, U):

        ellist = self.ELEMLIST
        nelem = len(ellist)

        print("--------------------------------------------------------")
        print("Element response:")
        print("--------------------------------------------------------")

        for i in range(nelem):
            id, nodes, ug = self.localize(i, U)
            el = self.ELEMLIST[i]
            el.getResponse(nodes)

    # -------------------------------------------------
    # Print nodal displacements
    # -------------------------------------------------
    def printDOF(self, U):

        nnodes, ndim = self.NODES.shape
        dummy, ndofn = self.CONSTRAINTS.shape

        print("--------------------------------------------------------")
        line = ("Node")
        for j in range(ndofn):
            line += ("    DOF %-3d " % (j+1))

        print(line)
        print("--------------------------------------------------------")
        for i in range(nnodes):
            line = ("%3d " % (i+1))
            for j in range(ndofn):
                dof = self.DOF[i,j]
                line += (" %10.5f " % U[dof])
            print(line)

    # -------------------------------------------------
    # Get data from the model to prepare for the analysis
    # -------------------------------------------------
    def getData(self):

        ndof = self.nfree
        nnodes = np.size(self.DOF,0)
        numdofnod = np.size(self.DOF,1)
        ntot = nnodes * numdofnod

        F = np.zeros(ntot)
        if np.size(self.NODELOAD, 1) > 0:
            for i in range(nnodes):
                id = self.DOF[i,:]
                F[id] = F[id] + self.NODELOAD[i,:]

        M = np.zeros(ntot)
        if np.size(self.NODEMASS, 1) > 0:
            for i in range(nnodes):
                id = self.DOF[i,:]
                M[id] = M[id] + self.NODEMASS[i,:]
        M = np.diag(M)

        ellist = self.ELEMLIST

        return ndof, ntot, F, M, ellist

    # -------------------------------------------------
    # Get derivative of mass matrix for DDM calculations
    # -------------------------------------------------
    def getMassDerivative(self, dMin):

        nnodes = np.size(self.DOF, 0)
        numdofnod = np.size(self.DOF, 1)
        ntot = nnodes * numdofnod
        dM = np.zeros(ntot)
        if np.size(dMin, 1) > 0:
            for i in range(nnodes):
                id = self.DOF[i,:]
                dM[id] = dM[id] + dMin[i][:]
        dM = np.diag(dM)

        return dM

    # -------------------------------------------------
    # Plot model
    # -------------------------------------------------
    def plotModel(self):

        # Create plot
        plt.ion()
        plt.figure()
        plt.autoscale(True)
        plt.title("Nodes and Elements")

        # Find outer plot limits
        xmin = 0
        xmax = 0
        ymin = 0
        ymax = 0

        nnodes = np.size(self.NODES, 0)
        for i in range(nnodes):

            if self.NODES[i, 0] < xmin:
                xmin = self.NODES[i, 0]

            elif self.NODES[i, 0] > xmax:
                xmax = self.NODES[i, 0]

            elif self.NODES[i, 1] < ymin:
                ymin = self.NODES[i, 1]

            elif self.NODES[i, 1] > ymax:
                ymax = self.NODES[i, 1]

        xmin = xmin - (xmax - xmin) * 0.2
        ymin = ymin - (ymax - ymin) * 0.2
        xmax = xmax + (xmax - xmin) * 0.2
        ymax = ymax + (ymax - ymin) * 0.2

        # Check if the structure is 1D
        if (ymax - ymin) == 0:
            ymin_new = ymin
            ymax = ymin
            ymin = ymin_new

        elif (xmax - xmin) == 0:
            xmin_new = xmin
            xmax = xmin
            xmin = xmin_new

        # Make the plotting area square
        xrange = xmax - xmin
        yrange = ymax - ymin

        if xrange > yrange:
            ymin -= 0.5 * (xrange - yrange)
            ymax += 0.5 * (xrange - yrange)
        else:
            xmin -= 0.5 * (yrange - xrange)
            xmax += 0.5 * (yrange - xrange)

        xcoord = np.zeros(2)
        ycoord = np.zeros(2)

        nelem = len(self.ELEMENTS)
        for i in range(nelem):

            node1 = int(self.ELEMENTS[i][len(self.ELEMENTS[i][:])-2])
            node2 = int(self.ELEMENTS[i][len(self.ELEMENTS[i][:])-1])
            xcoord[0] = self.NODES[node1-1, 0]
            ycoord[0] = self.NODES[node1-1, 1]
            xcoord[1] = self.NODES[node2-1, 0]
            ycoord[1] = self.NODES[node2-1, 1]
            plt.plot(xcoord, ycoord, 'ko-')

        plt.axis([xmin, xmax, ymin, ymax])
        print('\n'"Click somewhere in the plot to continue...")
        plt.waitforbuttonpress()

    # -------------------------------------------------
    # Plot deformed model
    # -------------------------------------------------
    def plotDeformedModel(self, U):

        # Create plot
        plt.ion()
        plt.figure()
        plt.autoscale(True)

        # Find outer plot limits
        xmin = 0
        xmax = 0
        ymin = 0
        ymax = 0

        numNodes = np.size(self.NODES, 0)
        for i in range(numNodes):

            if self.NODES[i, 0] < xmin:
                xmin = self.NODES[i, 0]

            elif self.NODES[i, 0] > xmax:
                xmax = self.NODES[i, 0]

            elif self.NODES[i, 1] < ymin:
                ymin = self.NODES[i, 1]

            elif self.NODES[i, 1] > ymax:
                ymax = self.NODES[i, 1]

        xmin = xmin - (xmax - xmin) * 0.2
        ymin = ymin - (ymax - ymin) * 0.2
        xmax = xmax + (xmax - xmin) * 0.2
        ymax = ymax + (ymax - ymin) * 0.2

        # Find scaling factor for displacement
        x_disp_max = 0.0
        y_disp_max = 0.0

        for i in range(numNodes):

            if abs(U[self.DOF[i,0]]) > x_disp_max:
                x_disp_max = abs(U[self.DOF[i,0]])

            if abs(U[self.DOF[i,1]]) > y_disp_max:
                y_disp_max = abs(U[self.DOF[i,1]])

        charact_dim = np.sqrt((xmax - xmin) ** 2 + (ymax - ymin) ** 2)

        if x_disp_max == 0.0 and y_disp_max == 0.0:
            scale_factor = 0.0
        elif x_disp_max > y_disp_max:
            scale_factor = 0.05 * charact_dim / x_disp_max
        else:
            scale_factor = 0.05 * charact_dim / y_disp_max

        # Add max displacement margin around plot
        xmin = xmin - x_disp_max * scale_factor
        ymin = ymin - y_disp_max * scale_factor
        xmax = xmax + x_disp_max * scale_factor
        ymax = ymax + y_disp_max * scale_factor

        # Check if the structure is 1D
        if (ymax - ymin) == 0:
            ymin_new = ymin - y_disp_max * scale_factor * 1.2
            ymax = ymin + y_disp_max * scale_factor * 1.2
            ymin = ymin_new

        elif (xmax - xmin) == 0:
            xmin_new = xmin - x_disp_max * scale_factor * 1.2
            xmax = xmin + x_disp_max * scale_factor * 1.2
            xmin = xmin_new

        # Make the plotting area square
        xrange = xmax - xmin
        yrange = ymax - ymin

        if xrange > yrange:
            ymin -= 0.5 * (xrange - yrange)
            ymax += 0.5 * (xrange - yrange)
        else:
            xmin -= 0.5 * (yrange - xrange)
            xmax += 0.5 * (yrange - xrange)

        xcoord = np.zeros(2)
        ycoord = np.zeros(2)

        nelem = np.size(self.ELEMENTS, 0)
        for i in range(nelem):

            node1 = int(self.ELEMENTS[i][len(self.ELEMENTS[i][:])-2])
            node2 = int(self.ELEMENTS[i][len(self.ELEMENTS[i][:])-1])

            xcoord[0] = self.NODES[node1-1, 0] + U[self.DOF[node1-1, 0]] * scale_factor
            ycoord[0] = self.NODES[node1-1, 1] + U[self.DOF[node1-1, 1]] * scale_factor
            xcoord[1] = self.NODES[node2-1, 0] + U[self.DOF[node2-1, 0]] * scale_factor
            ycoord[1] = self.NODES[node2-1, 1] + U[self.DOF[node2-1, 1]] * scale_factor

            plt.plot(xcoord, ycoord, 'ko-')

        plt.axis([xmin, xmax, ymin, ymax])
        plt.title("Deformed Model (max. disp. = %.4f)" % (max(x_disp_max, y_disp_max)))
        print('\n'"Click somewhere in the plot to continue...")
        plt.waitforbuttonpress()
