# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

from G2MaterialBilinear import *
from G2MaterialPlasticity import *
from G2MaterialBoucWen import *

class rectangularSection():

    # -------------------------------------------------
    # Constructor
    # -------------------------------------------------
    def __init__(self, h, b, nfib, material):

        self.mat = material

        self.h = h
        self.b = b
        self.nfib = nfib

        # Storing fibers and positions
        self.fibers = []
        self.locations = []
        self.areas = []

        # Create web fibers
        fibthick = self.h / self.nfib
        area = fibthick * self.b
        yf = 0.5*self.h - 0.5*fibthick
        for i in range(self.nfib):
            if self.mat[0] == 'Bilinear':
                fiber = bilinearMaterial(self.mat)
            elif self.mat[0] == 'Plasticity':
                fiber = plasticityMaterial(self.mat)
            elif  self.mat[0] == 'BoucWen':
                fiber = boucWenMaterial(self.mat)
            else:
                print("Error: Wrong material given to rectangular section")
                import sys
                sys.exit()

            self.fibers.append(fiber)
            self.locations.append(yf)
            self.areas.append(area)
            yf = yf - fibthick

        # Container for section deformations
        self.vs = [0.0, 0.0]

        # Commit total deformations to accommodate element13
        self.trialUsTotal0 = 0.0
        self.trialUsTotal1 = 0.0
        self.committedUsTotal0 = 0.0
        self.committedUsTotal1 = 0.0


    # -------------------------------------------------
    # Set parameter
    # -------------------------------------------------
    def setParameter(self, parameter, value):

        if parameter == 'h':
            self.h = value
        elif parameter == 'b':
            self.b = value
        else:
            for i in range(self.nfib):
                self.fibers[i].setParameter(parameter, value)
            return

        # Storing fibers and positions
        self.fibers = []
        self.locations = []
        self.areas = []

        # Create web fibers
        fibthick = self.h / self.nfib
        area = fibthick * self.b
        yf = 0.5*self.h - 0.5*fibthick
        for i in range(self.nfib):
            if self.mat[0] == 'Bilinear':
                fiber = bilinearMaterial(self.mat)
            elif self.mat[0] == 'Plasticity':
                fiber = plasticityMaterial(self.mat)
            elif  self.mat[0] == 'BoucWen':
                fiber = boucWenMaterial(self.mat)
            else:
                print("Error: Wrong material given to wide-flange section")
                import sys
                sys.exit()

            self.fibers.append(fiber)
            self.locations.append(yf)
            self.areas.append(area)
            yf = yf - fibthick


    # -------------------------------------------------
    # State determination
    # -------------------------------------------------
    def state(self, us):

        # Store total deformations to accommodate element13 & total strain materials
        self.trialUsTotal0 = self.committedUsTotal0 + us[0, 1]
        self.trialUsTotal1 = self.committedUsTotal1 + us[1, 1]

        Ks = np.zeros((2, 2))
        Fs = np.zeros(2)

        for i in range(self.nfib):

            # Fiber information
            fiber = self.fibers[i]
            area = self.areas[i]
            loc = self.locations[i]

            # Transformation
            Tms = np.array([-loc, 1])

            # Change in strain
            epsilon = Tms.dot(us)

            # Material state determination for fiber
            sigma, tangent = fiber.state(epsilon)

            # Section integration
            Ks = Ks + np.outer(Tms, Tms) * area * tangent
            Fs = Fs + Tms * area * sigma

        return Fs, Ks


    # -------------------------------------------------
    # DDM sensitivity analysis
    # -------------------------------------------------
    def stateDerivative(self, us, ddmParameter, ddmIndex, ddmIsHere, dKflag='none'):

        dKs = np.zeros((2,2))
        dFs = np.zeros(2)
        dKsdus = np.zeros((2,2,2))

        if ddmParameter in ['h', 'b'] and ddmIsHere:

            if ddmParameter == 'h':
                dFibthick = 1.0 / self.nfib
                dArea = dFibthick * self.b
                dyf = 0.5 - 0.5 * dFibthick

            elif ddmParameter == 'b':
                fibthick = self.h / self.nfib
                dFibthick = 0.0
                dArea = fibthick
                dyf = 0.0

            for i in range(self.nfib):
                fiber = self.fibers[i]
                area = self.areas[i]
                loc = self.locations[i]
                Tms = np.array([-loc, 1])
                epsilon = Tms.dot(us)
                sigma, E = fiber.state(epsilon)
                dSigma, dE, dKmdum = fiber.stateDerivative(epsilon, ddmParameter, ddmIndex, False, dKflag)
                dTms = np.array([-dyf, 0])
                dEpsilon = dTms.dot(us)
                dFs = dFs + dTms * area * sigma + Tms * dArea * sigma + Tms * area * E * dEpsilon[0] + Tms * area * dSigma
                if dKflag == 'Initial':
                    E = fiber.initialStiffness()
                dKs = dKs + np.outer(dTms, Tms) * area * E + np.outer(Tms, dTms) * area * E + np.outer(Tms, Tms) * dArea * E + np.outer(Tms, Tms) * area * (dKmdum * dEpsilon[0] + dE)
                A, B, C = np.ix_(Tms, Tms, Tms)
                dKsdus += np.multiply(dKmdum * area, A * B * C)

                dyf = dyf - dFibthick
        else:

            for i in range(self.nfib):

                fiber = self.fibers[i]
                area = self.areas[i]
                loc = self.locations[i]
                Tms = np.array([-loc, 1])
                epsilon = Tms.dot(us)
                dSigma, dE, dKmdum = fiber.stateDerivative(epsilon, ddmParameter, ddmIndex, ddmIsHere, dKflag)
                dKs = dKs + np.outer(Tms, Tms) * area * dE
                dFs = dFs + Tms * area * dSigma
                A, B, C = np.ix_(Tms, Tms, Tms)
                dKsdus += np.multiply(dKmdum * area, A * B * C)

        return dFs, dKs, dKsdus


    # -------------------------------------------------
    # Equilibrium
    # -------------------------------------------------
    def equilibrium(self, Fs):

        # Set iteration controls
        maxiter = 100
        tol = 1e-5

        # Initialize trial displacement vector for iteration
        usTrial = np.zeros(2)

        # Iterate on section equilibrium
        for j in range(maxiter):

            # This method is called by element13, which works with incremental deformations.
            # That means vsTrial is the second column of the U-matrix. Keep a history variable
            # to give total deformation to materials that work with that.
            self.trialUsTotal0 = self.committedUsTotal0 + usTrial[0]
            self.trialUsTotal1 = self.committedUsTotal1 + usTrial[1]
            FsTrial, Ks = self.state(np.array([[self.trialUsTotal0, usTrial[0], 0.0],
                                               [self.trialUsTotal1, usTrial[1], 0.0]]))

            FsResidual = Fs - FsTrial

            residualNorm = np.linalg.norm(FsResidual)

            if residualNorm < tol:
                break

            usTrial = np.linalg.solve(Ks, FsResidual) + usTrial

        if residualNorm >= tol:
            print("ERROR: Equilibrium iterations in section failed with residual", residualNorm)
            import sys
            sys.exit()

        return usTrial, Ks


    # -------------------------------------------------
    # Commit
    # -------------------------------------------------
    def commit(self):

        # Commit total deformation to accommodate element13
        self.committedUsTotal0 = self.trialUsTotal0
        self.committedUsTotal1 = self.trialUsTotal1

        # Commit material states
        for i in range(self.nfib):

            fiber = self.fibers[i]
            fiber.commit()


    # -------------------------------------------------
    # Commit sensitivity history variables
    # -------------------------------------------------
    def commitSensitivity(self, us, ddmus, ddmParameter, ddmIndex, ddmIsHere):

        if ddmParameter == 'h':

            dFibthick = 1.0 / self.nfib
            dyf = 0.5 - 0.5 * dFibthick

            for i in range(self.nfib):
                fiber = self.fibers[i]
                loc = self.locations[i]
                Tms = np.array([-loc, 1])
                epsilon = Tms.dot(us)
                dTms = np.array([-dyf, 0])
                ddmepsilon = dTms.dot(us[:, 0]) + Tms.dot(ddmus)
                fiber.commitSensitivity(epsilon, ddmepsilon, ddmParameter, ddmIndex, False)

                dyf = dyf - dFibthick
        else:

            for i in range(self.nfib):
                fiber = self.fibers[i]
                loc = self.locations[i]
                Tms = np.array([-loc, 1])
                epsilon = Tms.dot(us)
                ddmepsilon = Tms.dot(ddmus)
                fiber.commitSensitivity(epsilon, ddmepsilon, ddmParameter, ddmIndex, ddmIsHere)


    # -------------------------------------------------
    # Print response
    # -------------------------------------------------
    def getResponse(self):

        Fs = np.zeros(2)

        # Loop over fibers
        for i in range(self.nfib):

            # Fiber information
            fiber = self.fibers[i]
            area = self.areas[i]
            loc = self.locations[i]

            # Transformation
            Tms = np.array([-loc, 1])

            # Material state determination for fiber
            sigma = fiber.getResponse()

            # Section integration
            Fs = Fs + Tms * area * sigma

        return Fs
