# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# This file solves in Python the problem solved by Mathematica in the file
# posted nearby, entitled Computational Plastic Capacity Analysis for a Frame
# However, the code presented here is generic, and may be altered to solve
# other problems. Be especially cautious when defining the DOFs vector; that
# is where axial deformations are neglected and boundary conditions defined
# ------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import linprog

# Input values in m
H = 3.5
L = 7.0

# Nodal coordinates
COORDS = [[0.0, 0.0],
          [0.0, H],
          [L/2, H],
          [L,   H],
          [L,   0.0]]

# Connect nodal DOFs to final structural DOFs (to remove axial deformations; also introduces boundary conditions)
DOFs = [[0, 0, 0],
        [1, 0, 2],
        [1, 3, 4],
        [1, 0, 5],
        [0, 0, 0]]

# Check input
if np.size(COORDS, 0) != np.size(DOFs, 0):
    print('\n'"The number of nodes in COORDS and DOFs must be the same.")
    import sys
    sys.exit()

# Place reference loads onto final structural DOFs
LOADS = [[1.0,  0.0, -2.0, 0.0, 0.0]]

# Check input
if np.size(LOADS, 1) != np.max(DOFs):
    print('\n'"The number of loads and number of final DOFs must be the same.")
    import sys
    sys.exit()

# Connect elements to nodes
ELEMENTS = [[1, 2],
            [2, 3],
            [3, 4],
            [5, 4]]

# Define Basic to Local transformation
def Tbl(L):
    return [[-1.0/L, 1.0, 1/L, 0.0],
            [-1.0/L, 0.0, 1/L, 1.0]]

# Define Local to Global transformation
def Tlg(sin, cos):
    return [[-sin, cos, 0.0,  0.0, 0.0, 0.0],
            [ 0.0, 0.0, 1.0,  0.0, 0.0, 0.0],
            [ 0.0, 0.0, 0.0, -sin, cos, 0.0],
            [ 0.0, 0.0, 0.0,  0.0, 0.0, 1.0]]

# Establish Tgf for each element, and stack them
for i in range(np.size(ELEMENTS, 0)):

    # Element length and orientation
    x1 = COORDS[ELEMENTS[i][0]-1][0]
    y1 = COORDS[ELEMENTS[i][0]-1][1]
    x2 = COORDS[ELEMENTS[i][1]-1][0]
    y2 = COORDS[ELEMENTS[i][1]-1][1]
    dx = [x2-x1, y2-y1]
    L = np.sqrt(np.dot(dx, dx))
    cos = dx[0]/L
    sin = dx[1]/L

    # Transformation from global to final
    Tgf = np.zeros((6, np.max(DOFs)))
    for j in range(3):
        if DOFs[ELEMENTS[i][0]-1][j] != 0:
            Tgf[j, DOFs[ELEMENTS[i][0]-1][j]-1] = 1
    for j in range(3):
        if DOFs[ELEMENTS[i][1]-1][j] != 0:
            Tgf[j+3, DOFs[ELEMENTS[i][1]-1][j]-1] = 1

    # Calculate Tbf = Tbl * Tlg * Tgf
    Tbf = np.dot(np.dot(Tbl(L), Tlg(sin,cos)), Tgf)

    # Stack those matrices in a unique way
    if i == 0:
        TbfStacked = np.copy(Tbf)
    else:
        TbfStacked = np.concatenate((TbfStacked, Tbf), axis=0)

# Work with the negative transpose from here, and place the load vector in front,
# in order for the load factor to be the first unknown in the LP problem
A = np.concatenate((np.transpose(LOADS), -np.transpose(TbfStacked)), axis=1)

# The equality constraint (equilibrium) has a zero right-hand side
b = np.zeros(np.max(DOFs))

# Define the vector that picks the first unknown as the variable to maximize
c = np.concatenate(([-1], np.zeros(2*np.size(ELEMENTS, 0))))

# The equality bounds are set to plus/minus unity, as a reference plastic moment
bounds = np.array([[-1e6, 1e6]])
for i in range(2*np.size(ELEMENTS, 0)):
    bounds = np.concatenate((bounds, [[-1, 1]]), axis=0)

results = linprog(c=c, A_eq=A, b_eq=b, bounds=bounds, method='highs-ds')

if results.status == 0:
    print('\n'"Found a solution to the LP problem.")
else:
    print('\n'"Could not find a solution to the LP problem.")
    import sys
    sys.exit()

# Plot the structure with plastic hinges as dots; assemble stiffness matrix at the same time
plt.ion()
plt.figure()
plt.title(r"Plastic Mechanism ($\mathbf{F}_u=\lambda\cdot \mathbf{F}_{ref}$ with $\lambda$=%.5f$\cdot M_u)$" % (results.x[0]))
plt.grid(False)
plt.autoscale(True)
Kf = np.zeros((np.max(DOFs), np.max(DOFs)))
hingeDistanceFromNode = 0.03
for i in range(np.size(ELEMENTS, 0)):

    # Record the plastic hinge state of the element
    yieldEnd1 = False
    yieldEnd2 = False

    # Nodal coordinates
    x1 = COORDS[ELEMENTS[i][0]-1][0]
    y1 = COORDS[ELEMENTS[i][0]-1][1]
    x2 = COORDS[ELEMENTS[i][1]-1][0]
    y2 = COORDS[ELEMENTS[i][1]-1][1]

    # Plot Node 1 if there is plastic hinge
    if np.abs(results.x[2*i+1]) == 1:
        plt.plot(x1+(x2-x1)*hingeDistanceFromNode, y1+(y2-y1)*hingeDistanceFromNode, 'ko', markersize=5.0)
        yieldEnd1 = True

    # Plot Node 2 if there is plastic hinge
    if np.abs(results.x[2*i+2]) == 1:
        plt.plot(x2+(x1-x2)*hingeDistanceFromNode, y2+(y1-y2)*hingeDistanceFromNode, 'ko', markersize=5.0)
        yieldEnd2 = True

    # Plot the line between Node 1 and Node 2
    plt.plot([x1, x2], [y1, y2], 'k-', linewidth=1.0)

    # Basic stiffness matrix
    EI = 1
    if not yieldEnd1 and not yieldEnd2:
        Kb = [[4*EI/L, 2*EI/L],
              [2*EI/L, 4*EI/L]]
    elif yieldEnd1 and not yieldEnd2:
        Kb = [[0.0, 0.0],
              [0.0, 3*EI/L]]
    elif not yieldEnd1 and yieldEnd2:
        Kb = [[3*EI/L, 0.0],
              [0.0,    0.0]]
    elif yieldEnd1 and yieldEnd2:
        Kb = [[0.0, 0.0],
              [0.0, 0.0]]

    # Element length and direction cosines
    dx = [x2-x1, y2-y1]
    L = np.sqrt(np.dot(dx, dx))
    cos = dx[0]/L
    sin = dx[1]/L

    # Transformation from global to final
    Tgf = np.zeros((6, np.max(DOFs)))
    for j in range(3):
        if DOFs[ELEMENTS[i][0]-1][j] != 0:
            Tgf[j, DOFs[ELEMENTS[i][0]-1][j]-1] = 1
    for j in range(3):
        if DOFs[ELEMENTS[i][1]-1][j] != 0:
            Tgf[j+3, DOFs[ELEMENTS[i][1]-1][j]-1] = 1

    # Tbf = Tbl * Tlg * Tgf
    Tbf = np.dot(np.dot(Tbl(L), Tlg(sin,cos)), Tgf)

    # Kf = Tbf^T * Kb * Tbf
    Kf += np.dot(np.dot(np.transpose(Tbf), Kb), Tbf)

# Eigenvalues and vectors of the stiffness matrix
eigenvalues, eigenvectors = np.linalg.eig(Kf)

# Print that information
print('\n'"The eigenvalues of K are (zero means mechanism):", eigenvalues)
print('\n'"The corresponding eigenvectors are these rows:")
for i in range(len(eigenvalues)):
    print(eigenvectors[i])

# Print the displaced shape associated with the first zero eigenvalue with non-zero displacement DOF
DOFs = np.array(DOFs)
dispDOFs = np.array(DOFs[:,:2])
nonZeroEntries = np.nonzero(dispDOFs)
uniqueDispDOFs = np.unique(dispDOFs[nonZeroEntries])
foundShape = False
for i in range(len(eigenvalues)):
    if foundShape:
        break
    if np.abs(eigenvalues[i]) < 1e-6:
        for j in range(len(uniqueDispDOFs)):
            eigenvectorAtDispDOFs = eigenvectors[np.subtract(uniqueDispDOFs, 1), i]
            if np.sum(np.abs(eigenvectorAtDispDOFs)) > 0.0:
                shape = eigenvectors[:, i]
                foundShape = True

# Forget scaling for now and simply add eigenvector to undeformed shape
if foundShape:
    deformedCOORDS = np.copy(COORDS)
    for i in range(np.size(COORDS, 0)):
        if DOFs[i, 0] != 0:
            deformedCOORDS[i, 0] -= shape[int(DOFs[i, 0])-1]
        if DOFs[i, 0] != 0:
            deformedCOORDS[i, 1] -= shape[int(DOFs[i, 1])-1]
else:
    print('\n'"Did not find a proper shape from solving the eigenvalue problem for the stiffness matrix.")

# Add that to the existing plot
for i in range(np.size(ELEMENTS, 0)):

    # Node 1 of the element, with or w/o plastic hinge
    x1 = deformedCOORDS[ELEMENTS[i][0]-1][0]
    y1 = deformedCOORDS[ELEMENTS[i][0]-1][1]

    # Node 2 of the element, with or w/o plastic hinge
    x2 = deformedCOORDS[ELEMENTS[i][1]-1][0]
    y2 = deformedCOORDS[ELEMENTS[i][1]-1][1]

    # Plot the line between Node 1 and Node 2
    plt.plot([x1, x2], [y1, y2], '--', color="gray", linewidth=1.0)

plt.axis('equal')
plt.savefig('MechanismScreenshot.pdf', format='pdf')
print('\n'"Click somewhere in the plot to continue...")
plt.waitforbuttonpress()
