# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt

# Define a 2-dof correction function
def correction(t, duration, residualDisp, residualVel):

    if t <= 0 or t >= duration:
        return 0.0
    else:
        amp1 = (np.pi * residualVel)/(2 * duration)
        amp2 = (np.pi * (2 * residualDisp - duration * residualVel))/duration**2
        return amp1 * np.sin(np.pi * t / duration) + amp2 * np.sin(2*np.pi * t / duration)

# Trapezoidal integration
def trapezoidalRecordIntegration(delta_t, record):

    integrated = [0]
    for i in range(1, len(record)):
        integrated.append(integrated[i-1] + delta_t * 0.5 * (record[i-1]+record[i]))

    return integrated

# The main function in this file
def removeResiduals(delta_t, fileIn, fileOut, plotFlag=True):

    # Load the time series
    acc = []
    f = open(fileIn, "r")
    lines = f.readlines()
    for oneline in lines:
        splitline = oneline.split()
        for j in range(len(splitline)):
            value = float(splitline[j])
            acc.append(value)
    f.close()

    # Calculate duration and create time axis
    numPoints = len(acc)
    duration = delta_t * numPoints
    t = np.linspace(0, duration, numPoints)

    # Integrate to obtain velocity and displacement
    vel = trapezoidalRecordIntegration(delta_t, acc)
    disp = trapezoidalRecordIntegration(delta_t, vel)

    # Determine residuals
    residualVelocity = vel[numPoints-1]
    residualDisplacement = disp[numPoints-1]

    # Print residuals
    if plotFlag:
        print('\n'"Original residual ground velocity %.6f and displacement %.6f" % (residualVelocity, residualDisplacement))

    # Plot the correction function
    corrFcn = []
    for i in range(numPoints):
        corrFcn.append(correction(t[i], duration, residualDisplacement, residualVelocity))
    if plotFlag:
        plt.ion()
        plt.figure()
        plt.plot(t, corrFcn, 'k-')
        plt.xlabel("Time")
        plt.title("Acceleration Correction")
        print('\n'"Click somewhere in the plot to continue...")
        plt.waitforbuttonpress()

    # Produce new acceleration time histories
    for i in range(numPoints):
        acc[i] -= correction(t[i], duration, residualDisplacement, residualVelocity)

    # Integrate to obtain velocity and displacement
    vel = trapezoidalRecordIntegration(delta_t, acc)
    disp = trapezoidalRecordIntegration(delta_t, vel)

    # Print residuals
    if plotFlag:
        print('\n'"New residual ground velocity %.6f and displacement %.6f" % (vel[numPoints-1], disp[numPoints-1]))

    # Plot
    if plotFlag:
        plt.ion()
        fig1 = plt.figure()
        plt.axis('off')
        plt.title('Time Series Without Residuals')

        axes = fig1.add_subplot(311)
        axes.get_xaxis().set_visible(False)
        plt.plot(t, acc, 'k-', linewidth=1.0)
        plt.ylabel('$\ddot{u}_g$')

        axes = fig1.add_subplot(312)
        axes.get_xaxis().set_visible(False)
        plt.plot(t, vel, 'k-', linewidth=1.0)
        plt.ylabel('$\dot{u}_g$')

        axes = fig1.add_subplot(313)
        plt.plot(t, disp, 'k-', linewidth=1.0)
        plt.ylabel('${u}_g$')
        plt.xlabel('Time [sec]')

        print('\n'"Click somewhere in the plot to continue...")
        plt.waitforbuttonpress()

    # Print the ground motions to file
    outFileID = open(fileOut, 'w')
    for step in range(numPoints):
        line = ("%12.6E" % acc[step])
        line += '\n'
        outFileID.write(line)
    outFileID.close()
