# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import numpy as np

# ------------------------------------------------------------------------
# Kl AND Tlg FOR TRUSS
# ------------------------------------------------------------------------
def getTrussStiffness2D(E, A, L, delta_x, delta_y, N):

    # Local stiffness matrix
    Kl = np.array([(E*A/L,  0, -E*A/L, 0),
                   (0,      0,   0,    0),
                   (-E*A/L, 0, E*A/L,  0),
                   (0,      0,   0,    0)])

    # Geometric stiffness matrix
    Kgeometric = np.array([(0,  0,  0,   0),
                          (0,  N/L, 0, -N/L),
                          (0,   0,  0,   0),
                          (0, -N/L, 0, N/L)])

    # Transformation matrix Tlg
    if delta_x == 0:
        if delta_y < 0:
            theta = -np.pi/2.0
        else:
            theta = np.pi/2.0
    else:
        theta = np.arctan(delta_y/delta_x)

    c = np.cos(theta)
    s = np.sin(theta)
    Tlg = np.array([(c,  s, 0,  0),
                    (-s, c, 0,  0),
                    (0,  0, c,  s),
                    (0,  0, -s, c)])

    return [Kl, Kgeometric, Tlg]

# ------------------------------------------------------------------------
# Kl AND Tlg FOR FRAME
# ------------------------------------------------------------------------
def getFrameStiffness2D(E, nu, I, A, beta, L, delta_x, delta_y, N):

    # Calculate alpha for shear deformation (assume beta=0 means user doesn't want this)
    if beta > 0.0:
        G = E / (2.0 * (1 + nu))
        alpha = 12.0 * E * I / (G * beta * A * L**2)
    else:
        alpha = 0

    # Local stiffness matrix
    Kl = np.array([(E*A/L,   0,            0,         -E*A/L, 0,            0),
                   (0,       12*E*I/L**3, -6*E*I/L**2, 0,    -12*E*I/L**3, -6*E*I/L**2),
                   (0,      -6*E*I/L**2,   (4+alpha)*E*I/L,    0,     6*E*I/L**2,   (2-alpha)*E*I/L),
                   (-E*A/L,  0,            0,          E*A/L, 0,            0),
                   (0,      -12*E*I/L**3,  6*E*I/L**2, 0,     12*E*I/L**3,  6*E*I/L**2),
                   (0,      -6*E*I/L**2,   (2-alpha)*E*I/L,    0,     6*E*I/L**2,   (4+alpha)*E*I/L)])

    Kl = np.multiply(Kl, 1.0/(1.0 + alpha))

    # Geometric stiffness matrix
    Kgeometric = np.array([(0,   0,        0,     0,    0,         0),
                   (0,  6/(5*L), -1/10,   0, -6/(5*L), -1/10),
                   (0, -1/10,     2*L/15, 0,  1/10,    -L/30),
                   (0,   0,        0,     0,     0,        0),
                   (0, -6/(5*L),  1/10,   0,  6/(5*L),  1/10),
                   (0, -1/10,    -L/30,   0,  1/10,     2*L/15)])

    Kgeometric = np.multiply(Kgeometric, N)

    # Transformation matrix Tlg
    if delta_x == 0:
        if delta_y < 0:
            theta = -np.pi/2.0
        else:
            theta = np.pi/2.0
    else:
        theta = np.arctan(delta_y/delta_x)

    c = np.cos(theta)
    s = np.sin(theta)
    Tlg = np.array([(c, s, 0,  0, 0, 0),
                    (-s, c, 0,  0, 0, 0),
                    (0,  0, 1,  0, 0, 0),
                    (0,  0, 0,  c, s, 0),
                    (0,  0, 0, -s, c, 0),
                    (0,  0, 0,  0, 0, 1)])

    return [Kl, Kgeometric, Tlg]

# ------------------------------------------------------------------------
# Kg FOR QUAD4
# ------------------------------------------------------------------------
def getQuad4Stiffness2D(E, nu, t, ps, x, y):

    # Set material stiffness
    D = np.zeros((3, 3))
    if ps == 1:  # Plane stress

        factor = E / (1.0-nu * nu)

        D[0, 0] = factor
        D[1, 0] = nu * factor
        D[2, 0] = 0.0
        D[0, 1] = nu * factor
        D[0, 2] = 0.0
        D[1, 1] = factor
        D[1, 2] = 0.0
        D[2, 1] = 0.0
        D[2, 2] = 0.5 * factor * (1.0 - nu)

    else:  # Plane strain

        factor = E / ((1.0+nu) * (1.0-2.0 * nu))

        D[0, 0] = factor * (1.0-nu)
        D[1, 0] = nu * factor
        D[2, 0] = 0.0
        D[0, 1] = nu * factor
        D[0, 2] = 0.0
        D[1, 1] = factor * (1.0 - nu)
        D[1, 2] = 0.0
        D[2, 1] = 0.0
        D[2, 2] = 0.5 * factor * (1.0 - 2.0 * nu)

    # Set quadrature rule
    order = 2
    integrationPoints = np.array([-0.577350269189626, 0.577350269189626])
    integrationWeights = np.array([1, 1])

    # Initialize loop variables
    N = np.zeros(4)
    dNdxi = np.zeros(4)
    dNdeta = np.zeros(4)
    dNdx = np.zeros(4)
    dNdy = np.zeros(4)
    iJac = np.zeros((2, 2))
    B = np.zeros((3, 8))
    Kg = np.zeros((8, 8))

    # Loop over integration points
    for k in range(order):
        for l in range(order):

            # Re-initialize the Jacobian matrix at each integration points
            jacobianMatrix = np.zeros((2, 2))

            # Get location of integeration points
            xi = integrationPoints[k]
            eta = integrationPoints[l]

            # Get shape functions, derivative and Jacobian at this quadrature point
            factor = 0.25

            N[0] = factor * (1.0-xi) * (1.0-eta)
            N[1] = factor * (1.0+xi) * (1.0-eta)
            N[2] = factor * (1.0+xi) * (1.0+eta)
            N[3] = factor * (1.0-xi) * (1.0+eta)

            dNdxi[0] = factor * -1.0 * (1.0-eta)
            dNdxi[1] = factor * 1.0 * (1.0-eta)
            dNdxi[2] = factor * 1.0 * (1.0+eta)
            dNdxi[3] = factor * -1.0 * (1.0+eta)

            dNdeta[0] = factor * (1.0-xi) * -1.0
            dNdeta[1] = factor * (1.0+xi) * -1.0
            dNdeta[2] = factor * (1.0+xi) * 1.0
            dNdeta[3] = factor * (1.0-xi) * 1.0

            for m in range(4):
                jacobianMatrix[0, 0] = jacobianMatrix[0, 0] + dNdxi[m] * x[m]
                jacobianMatrix[1, 0] = jacobianMatrix[1, 0] + dNdxi[m] * y[m]
                jacobianMatrix[0, 1] = jacobianMatrix[0, 1] + dNdeta[m] * x[m]
                jacobianMatrix[1, 1] = jacobianMatrix[1, 1] + dNdeta[m] * y[m]

            # Determine the Jacobian determinant
            jacDet = jacobianMatrix[0, 0] * jacobianMatrix[1, 1] - jacobianMatrix[0, 1] * jacobianMatrix[1, 0]

            # Determine the inverse of the Jacobian matrix
            iJac[0, 0] = 1.0 / jacDet * jacobianMatrix[1, 1]
            iJac[1, 1] = 1.0 / jacDet * jacobianMatrix[0, 0]
            iJac[0, 1] = -1.0 / jacDet * jacobianMatrix[0, 1]
            iJac[1, 0] = -1.0 / jacDet * jacobianMatrix[1, 0]

            # Determine components of the B-matrix
            for m in range(4):
                dNdx[m] = dNdxi[m] * iJac[0, 0] + dNdeta[m] * iJac[1, 0]
                dNdy[m] = dNdxi[m] * iJac[0, 1] + dNdeta[m] * iJac[1, 1]

            # Set the strain-displacement matrix B
            for m in range(4):
                B[0][(2 * m)] = dNdx[m]
                B[1][(2 * m)] = 0.0
                B[2][(2 * m)] = dNdy[m]

                B[0][(2 * m + 1)] = 0.0
                B[1][(2 * m + 1)] = dNdy[m]
                B[2][(2 * m + 1)] = dNdx[m]

            # Pre-compute Jacobian times quadrature weight (volume change)
            dvol = jacDet * t * integrationWeights[k] * integrationWeights[l]

            # Compute 8x8 stiffness matrix as B ^ T * D * B * dvol
            Kg += np.multiply((np.transpose(B).dot(D)).dot(B), dvol)

    return Kg

# ------------------------------------------------------------------------
# STIFFNESS MATRIX FROM "GLOBAL" TO "ALL"
# ------------------------------------------------------------------------
def assembleStiffnessMatrix(Ka, Kg, nodeList, numDOFsPerNodeInElement, numDOFsPerNodeInStructure):

    numNodesInElement = len(nodeList)

    for j in range(numNodesInElement):

        for k in range(j + 1):

            for m in range(numDOFsPerNodeInElement):

                for n in range(numDOFsPerNodeInElement):

                    Ka[(nodeList[j] - 1) * numDOFsPerNodeInStructure + m, (
                            nodeList[k] - 1) * numDOFsPerNodeInStructure + n] += Kg[
                        j * numDOFsPerNodeInElement + m, k * numDOFsPerNodeInElement + n]

                    if j != k:
                        Ka[(nodeList[k] - 1) * numDOFsPerNodeInStructure + n, (
                                nodeList[j] - 1) * numDOFsPerNodeInStructure + m] += Kg[
                            j * numDOFsPerNodeInElement + m, k * numDOFsPerNodeInElement + n]


    # This could have been done, less efficiently, by using Tga, like this:
    # Tga = np.zeros((numDOFsInElement, totalNumDOFs))
    # for j in range(nodeList.size):
    #     for k in range(numDOFsPerNodeInElement):
    #         Tga[j * numDOFsPerNodeInElement + k, (nodeList[j] - 1) * numDOFsPerNodeInStructure + k] = 1
    # Ka += (np.transpose(Tga).dot(Kg)).dot(Tga)

    return Ka

# ------------------------------------------------------------------------
# LOAD VECTOR FROM "GLOBAL" TO "ALL"
# ------------------------------------------------------------------------
def assembleLoadVector(Fa, Fg, nodeList, numDOFsPerNodeInElement, numDOFsPerNodeInStructure):

    numNodesInElement = len(nodeList)

    for j in range(numNodesInElement):

        for m in range(numDOFsPerNodeInElement):

            Fa[(nodeList[j] - 1) * numDOFsPerNodeInStructure + m] += Fg[j * numDOFsPerNodeInElement + m]

    return Fa
