# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------

import numpy as np

# ------------------------------------------------------------------------
# INPUT
# ------------------------------------------------------------------------

q = 300
E = 60000
nu = 0.1
h = 150

# Mesh parameters
plateLength = 9000
plateWidth = 1500
nelx = 2
nely = 2
xinterval = plateLength / nelx
yinterval = plateWidth / nely

# Nodes (x, y)
nodes = np.zeros(((nelx + 1) * (nely + 1), 2))
counter = 0
for j in range(nely + 1):
    for i in range(nelx + 1):
        nodes[counter, 0] = i * xinterval
        nodes[counter, 1] = j * yinterval
        counter += 1

# Mindlin elements (node1, node2, node3, node3, E, nu, h)
mindlinElements = np.zeros((nelx * nely, 8))
counter = 0
for j in range(nely):
    for i in range(nelx):
        mindlinElements[counter, 0] = i + j * (nelx + 1) + 1
        mindlinElements[counter, 1] = i + j * (nelx + 1) + 2
        mindlinElements[counter, 2] = i + (j + 1) * (nelx + 1) + 2
        mindlinElements[counter, 3] = i + (j + 1) * (nelx + 1) + 1
        mindlinElements[counter, 4] = E
        mindlinElements[counter, 5] = nu
        mindlinElements[counter, 6] = h
        counter += 1

# Loads (node, value, direction, factor) [DOF ordering: 1-vertical, 2-bending about... 3-bending about...]
P = q * (plateLength/nelx) * (plateWidth/nely)
numNodes = nodes.shape[0]
loads = np.zeros((numNodes, 4))
counter = 0
for i in range(numNodes):
    loads[i, 0] = i+1
    loads[i, 1] = -P
    loads[i, 2] = 1
    loads[i, 3] = 1

# Boundary conditions (node, 0=free, 1=fixed)
constraints = np.zeros((2*(nely+1)+2*(nelx-1), 4))
counter = 0
for j in range(nely+1):
    constraints[counter, 0] = j * (nelx+1) + 1
    constraints[counter, 1] = 1
    constraints[counter, 2] = 1
    constraints[counter, 3] = 1
    counter += 1
    constraints[counter, 0] = j * (nelx+1) + 1 + nelx
    constraints[counter, 1] = 1
    constraints[counter, 2] = 1
    constraints[counter, 3] = 1
    counter += 1
for i in range(nelx-1):
    constraints[counter, 0] = i+2
    constraints[counter, 1] = 1
    constraints[counter, 2] = 1
    constraints[counter, 3] = 1
    counter += 1
    constraints[counter, 0] = (nely+1)*(nelx-1) + i + 3
    constraints[counter, 1] = 1
    constraints[counter, 2] = 1
    constraints[counter, 3] = 1
    counter += 1

# ------------------------------------------------------------------------
# EXTRACT INFORMATION
# ------------------------------------------------------------------------

# Number of nodes, elements, and DOFs
numDOFsPerNodeInStructure = 3
numNodes = nodes.shape[0]
totalNumDOFs = numDOFsPerNodeInStructure * numNodes
numNodesInElement = 4
numDOFsPerNodeInElement = 3
numDOFsInElement = numDOFsPerNodeInElement * numNodesInElement
numMindlinElements = mindlinElements.shape[0]

# ------------------------------------------------------------------------
# INITIALIZE
# ------------------------------------------------------------------------
Ka = np.zeros((totalNumDOFs, totalNumDOFs))
Fa = np.zeros(totalNumDOFs)


for i in range(numMindlinElements):

    # ------------------------------------------------------------------------
    # ASSEMBLE THE STIFFNESS MATRIX
    # ------------------------------------------------------------------------

    # Information about the element
    E = mindlinElements[i, 4]
    nu = mindlinElements[i, 5]
    h = mindlinElements[i, 6]

    # Information about nodes
    node1 = int(mindlinElements[i, 0])
    node2 = int(mindlinElements[i, 1])
    node3 = int(mindlinElements[i, 2])
    node4 = int(mindlinElements[i, 3])
    nodelist = np.array([node1, node2, node3, node4])
    nodalcoords = np.array([nodes[node1 - 1, 0], nodes[node1 - 1, 1], nodes[node2 - 1, 0], nodes[node2 - 1, 1],
                            nodes[node3 - 1, 0], nodes[node3 - 1, 1], nodes[node4 - 1, 0], nodes[node4 - 1, 1]])

    x = np.array([nodalcoords[0], nodalcoords[2], nodalcoords[4], nodalcoords[6]])
    y = np.array([nodalcoords[1], nodalcoords[3], nodalcoords[5], nodalcoords[7]])
    
    # Initial declarations
    Kquad = np.zeros((12, 12))
    B = np.zeros((5, 12))
    N = np.zeros(4)
    dNdxi = np.zeros(4)
    dNdeta = np.zeros(4)
    dNdx = np.zeros(4)
    dNdy = np.zeros(4)
    iJac = np.zeros((2, 2))
    
    # Set bending stiffness
    DD = np.zeros((5, 5))
    G = E / (2.0 * (1.0 + nu**2))
    D = E * h * h * h / (12.0 * (1.0 - nu * nu))
    kf = 0.833333
    
    DD[0, 0] = D
    DD[1, 0] = nu*D
    DD[2, 0] = 0.0
    DD[3, 0] = 0.0
    DD[4, 0] = 0.0
    
    DD[0, 1] = nu*D
    DD[1, 1] = D
    DD[2, 1] = 0.0
    DD[3, 1] = 0.0
    DD[4, 1] = 0.0
    
    DD[0, 2] = 0.0
    DD[1, 2] = 0.0
    DD[2, 2] = 0.5*D*(1.0-nu)
    DD[3, 2] = 0.0
    DD[4, 2] = 0.0
    
    DD[0][3] = 0.0
    DD[1][3] = 0.0
    DD[2][3] = 0.0
    DD[3][3] = kf * G * h
    DD[4][3] = 0.0
    
    DD[0, 4] = 0.0
    DD[1, 4] = 0.0
    DD[2, 4] = 0.0
    DD[3, 4] = 0.0
    DD[4, 4] = kf * G * h
    
    # Set quadrature rule
    order = 2
    integrationPoints = np.array([-0.577350269189626, 0.577350269189626])
    integrationWeights = np.array([1, 1])
    
    # Loop over integration points
    for k in range(order):
        for l in range(order):
    
            # Re-initialize the Jacobian matrix at each integration points
            jacobianMatrix = np.zeros((2, 2))
    
            # Get location of integeration points
            xi = integrationPoints[k]
            eta = integrationPoints[l]
    
            # Get shape functions, derivative and Jacobian at this quadrature point
            N[0] = 0.25 * (1.0-xi) * (1.0-eta)
            N[1] = 0.25 * (1.0+xi) * (1.0-eta)
            N[2] = 0.25 * (1.0+xi) * (1.0+eta)
            N[3] = 0.25 * (1.0-xi) * (1.0+eta)
    
            dNdxi[0] = 0.25 * (-1.0) * (1.0-eta)
            dNdxi[1] = 0.25 * (1.0) * (1.0-eta)
            dNdxi[2] = 0.25 * (1.0) * (1.0+eta)
            dNdxi[3] = 0.25 * (-1.0) * (1.0+eta)
    
            dNdeta[0] = 0.25 * (1.0-xi) * (-1.0)
            dNdeta[1] = 0.25 * (1.0+xi) * (-1.0)
            dNdeta[2] = 0.25 * (1.0+xi) * (1.0)
            dNdeta[3] = 0.25 * (1.0-xi) * (1.0)
    
            for m in range(4):
                jacobianMatrix[0, 0] += dNdxi[m] * x[m]       
                jacobianMatrix[1, 0] += dNdxi[m] * y[m]
    
                jacobianMatrix[0, 1] += dNdeta[m] * x[m]
                jacobianMatrix[1, 1] += dNdeta[m] * y[m]
    
    
            # Determine the Jacobian determinant
            jacDet = jacobianMatrix[0, 0] * jacobianMatrix[1, 1] - jacobianMatrix[0, 1] * jacobianMatrix[1, 0]

    
            # Determine the inverse of the Jacobian matrix
            iJac[0, 0] = 1.0/jacDet*jacobianMatrix[1, 1]
            iJac[1, 1] = 1.0/jacDet*jacobianMatrix[0, 0]
            iJac[0, 1] = -1.0/jacDet*jacobianMatrix[0, 1]
            iJac[1, 0] = -1.0/jacDet*jacobianMatrix[1, 0]

            # Determine the B-matrix
            for m in range(4):
                dNdx[m] = dNdxi[m] * iJac[0, 0] + dNdeta[m] * iJac[1, 0]
                dNdy[m] = dNdxi[m] * iJac[0, 1] + dNdeta[m] * iJac[1, 1]

            for m in range(4):
                B[0, (3*m)] = 0.0
                B[1, (3*m)] = 0.0
                B[2, (3*m)] = 0.0
                B[3, (3*m)] = -dNdy[m]
                B[4, (3*m)] = dNdx[m]

                B[0, (3*m+1)] = 0.0
                B[1, (3*m+1)] = -dNdy[m]
                B[2, (3*m+1)] = dNdx[m]
                B[3, (3*m+1)] =-N[m]
                B[4, (3*m+1)] = 0.0

                B[0, (3*m+2)] =-dNdx[m]
                B[1, (3*m+2)] = 0.0
                B[2, (3*m+2)] =dNdy[m]
                B[3, (3*m+2)] = 0.0
                B[4, (3*m+2)] = N[m]

                # Accumulate the stiffness matrix B^T * D * B * jacDet * weightI * weightJ
                Kquad += np.multiply(((np.transpose(B)).dot(DD)).dot(B), jacDet * integrationWeights[k] * integrationWeights[l])

        connTable = np.array([node1, node2, node3, node4])

        for j in range(numNodesInElement):

            for k in range(j + 1):

                for m in range(numDOFsPerNodeInElement):

                    for n in range(numDOFsPerNodeInElement):

                        Ka[(connTable[j] - 1) * numDOFsPerNodeInStructure + m, (
                                    connTable[k] - 1) * numDOFsPerNodeInStructure + n] \
                            += Kquad[j * numDOFsPerNodeInElement + m, k * numDOFsPerNodeInElement + n]

                        if j != k:
                            Ka[(connTable[k] - 1) * numDOFsPerNodeInStructure + n, (
                                        connTable[j] - 1) * numDOFsPerNodeInStructure + m] \
                                += Kquad[j * numDOFsPerNodeInElement + m, k * numDOFsPerNodeInElement + n]

    # ------------------------------------------------------------------------
    # ASSEMBLE LOAD VECTOR
    # ------------------------------------------------------------------------
    numLoads = loads.shape[0]

    for i in range(numLoads):
        Fa[(int(loads[i, 0])-1)*numDOFsPerNodeInStructure+(int(loads[i,  2])-1)] = loads[i, 1] * loads[i, 3]

    # ------------------------------------------------------------------------
    # DETECT LOCKED DOFs
    # ------------------------------------------------------------------------

    # DOFs locked by input
    dofStatus = np.zeros(totalNumDOFs)

    nfix = constraints.shape[0]
    for k in range(nfix):
        fixedNode = int(constraints[k, 0])
        for m in range(numDOFsPerNodeInStructure):
            if constraints[k, m+1] == 1:
                dofStatus[(fixedNode-1)*numDOFsPerNodeInStructure+m] = 1

    # DOFs without stiffness
    for i in range(totalNumDOFs):
        if Ka[i, i] == 0:
            dofStatus[i] = 1

    numFreeDOFs = int(totalNumDOFs - np.sum(dofStatus))

    # ------------------------------------------------------------------------
    # INTRODUCE BOUNDARY CONDITIONS BY CONNECTIVITY MATRIX
    # ------------------------------------------------------------------------
    Kf = np.zeros((numFreeDOFs, numFreeDOFs))
    Ff = np.zeros(numFreeDOFs)
    iCounter = 0
    jCounter = 0
    for i in range(totalNumDOFs):

        if dofStatus[i] == 0:

            jCounter = 0

            for j in range(i+1):

                if dofStatus[j] == 0:

                    Kf[iCounter, jCounter] = Ka[i, j]
                    Kf[jCounter, iCounter] = Ka[i, j]
                    jCounter += 1

            Ff[iCounter] = Fa[i]
            iCounter += 1

    # ------------------------------------------------------------------------
    # LINEAR SOLVER
    # ------------------------------------------------------------------------
    uf = np.linalg.solve(Kf, Ff)

    # ------------------------------------------------------------------------
    # GET ALL DISPLACEMENTS USING CONNECTIVITY MATRIX
    # ------------------------------------------------------------------------
    ua = np.zeros(totalNumDOFs)
    iCounter = 0
    for i in range(totalNumDOFs):
        if dofStatus[i] == 0:
            ua[i] = uf[iCounter]
            iCounter += 1

print("uf:", uf)