# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# ------------------------------------------------------------------------
# Notation:
# F(x) = objective function, or merit function, relevant in optimization, not in root-finding
# f(x) = dF/dx = derivative of the objective function, i.e., the function whose root is sought
# h(x) = df/dx = d^2F/dx^2 = second-order derivative, i.e., Hessian of the objective function
# ------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt
from FiniteDifference import *

def newtonLineSearch(F, startPoint, maxIterations, tolerance, plot):

    # Initiate plot if requested (plot = delay if positive)
    if plot > 0:

        plt.clf()
        plt.ion()
        plt.title('Newton Search')
        plt.grid(True)
        plt.autoscale(True)
        plt.ylabel('Derivative of Objective Function')
        plt.xlabel('Design Variable')
        xData = []
        fData = []

    # Start the loop
    counter = 0
    convergence = False
    x = startPoint
    while not convergence:

        # Increment counter
        counter += 1

        # Exit if we are at max number of iterations
        if counter > maxIterations:
            print("Newton search reached the maximum number of iterations without convergence")
            optimum = 0
            break

        # Initialize the search
        xPrevious = x

        # Evaluate the ingredients of the Newton fraction
        fValue = f(x, F)
        df = h(x, F)

        # Add point to the plot
        if plot > 0:
            xData.append(x)
            fData.append(fValue)
            plt.plot(xData, fData, 'ko', linewidth=1.0)
            plt.pause(plot)

        # Evaluate the Newton formula
        x = xPrevious - fValue / df

        # Output
        # print("At step", counter, "the design variable value is", x)

        # Check convergence
        if np.abs(x - xPrevious) < tolerance:
            convergence = True
            optimum = x

    # Output
    print('\n'"Newton search done after", counter, "steps with solution", optimum)

    # Hold the plot for a few seconds before proceeding
    if plot > 0:
        print('\n'"Click somewhere in the plot to continue...")
        plt.waitforbuttonpress()

    return optimum
