# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. Please be cautious when using
# this code; it may contain bugs and comes without warranty of any kind.
# The good parts of this particular code is from an email that Professor
# Michael Scott at Oregon State sent me, which is gratefully acknowledged.
# ------------------------------------------------------------------------

import numpy as np
import matplotlib.pyplot as plt

# System of springs
#    |                                                               |
#    |-------(Spring 1)--------| DOF 1 >>> |-------(Spring 2)--------|
#    |                                                               |
#    |                                                               | DOF 2 >>>
#    |                                                               |
#    |----------------------------(Spring 3)-------------------------|
#    |                                                               |


# Basic spring responses
def spring(u, uPeak, Fpeak):
    force = Fpeak * (3.0 * (u/uPeak) - (u/uPeak)**3)
    stiffness = Fpeak/uPeak * (3.0 - 3.0 * (u/uPeak)**2)
    return force, stiffness

def spring1(u):
    return spring(u, 1.5, 5.0)

def spring2(u):
    return spring(u, 3.0, 2.0)

def spring3(u):
    return spring(u, 1.5, 3.5)


# Plot the force-deformation relationships
u = np.linspace(0, 5.5, 50)
plt.ion()
fig, ax = plt.subplots()
ax.plot(u, (spring1(u))[0], 'k', label='Spring 1')
ax.plot(u, (spring2(u))[0], 'b', label='Spring 2')
ax.plot(u, (spring3(u))[0], 'r', label='Spring 3')
plt.ylim([-1, 10.5])
plt.xlabel("Displacement")
plt.ylabel("Force")
plt.title("Spring Responses")
legend = ax.legend(loc='upper right')
print('\n'"Click somewhere in the plot to continue...")
plt.waitforbuttonpress()


# Transformations from Basic to Final
Tbf1 = np.array([[ 1, 0]])
Tbf2 = np.array([[-1, 1]])
Tbf3 = np.array([[ 0, 1]])


# Analysis setup
lam = 0
controlDOF = 1
uoFinal = 2.5
duo = 0.2
tol = 1e-3
maxIterations = 50
numIncrements = int(uoFinal/duo)


# Load pattern
Fref = np.array([0, 10])


# Initial trial displacements
uf = np.array([0.0, 0.0])


# Loop over load steps [*** INCREMENTS ***]
t = 0
loadArray = [0]
disp1Array = [0]
disp2Array = [0]
for n in range (numIncrements):

    # Keep the user posted
    print('\n'"************ Starting increment %i ************" % (n+1))

    # Impose displacement at the control DOF
    uf[controlDOF] += duo

    # Corresponding Basic displacements
    ub1 = np.dot(Tbf1, uf)
    ub2 = np.dot(Tbf2, uf)
    ub3 = np.dot(Tbf3, uf)

    # Basic stiffnesses
    Kb1 = (spring1(ub1))[1]
    Kb2 = (spring2(ub2))[1]
    Kb3 = (spring3(ub3))[1]

    # Final stiffness matrix
    Kf = np.transpose(Tbf1 * Kb1).dot(Tbf1) \
         + np.transpose(Tbf2 * Kb2).dot(Tbf2) \
         + np.transpose(Tbf3 * Kb3).dot(Tbf3)

    # Solve for "tangent displacements" [*** T ***]
    ufT = np.linalg.solve(Kf, Fref)

    # Determine initial load increment
    dlambda = duo / ufT[controlDOF]
    lam += dlambda

    # Set corresponding load level
    Ff = lam * Fref

    # State determination for target displacements, starting with Basic forces
    Fb1 = (spring1(ub1))[0]
    Fb2 = (spring2(ub2))[0]
    Fb3 = (spring3(ub3))[0]

    # Final force vector
    tildeFf = np.transpose(Tbf1).dot(Fb1) \
              + np.transpose(Tbf2).dot(Fb2) \
              + np.transpose(Tbf3).dot(Fb3)

    # Residual vector and its norm
    Rf = tildeFf - Ff
    resNorm = np.linalg.norm(Rf)

    # Loop to force the residual to zero [*** ITERATIONS ***]
    i = 0
    iArray = [str(i)]
    resArray = [resNorm]
    while resNorm > tol and i < maxIterations:

        # Keep the user posted
        print("Starting iteration", i+1)

        # Determine "residual displacements" [*** R ***]
        dufR = np.linalg.solve(Kf, -Rf)

        # Determine load increment
        dlambda = -dufR[controlDOF] / ufT[controlDOF]
        lam += dlambda

        # Set corresponding load level and store it for plotting
        Ff = lam * Fref
        if i == 0:
            loadArray.append(Ff[1])
        else:
            loadArray[len(loadArray)-1] = Ff[1]

        # Update trial displacements
        duf = dufR + dlambda * ufT
        uf += duf

        # Store displacements for plotting purposes
        if i == 0:
            disp1Array.append(uf[0])
            disp2Array.append(uf[1])
        else:
            disp1Array[len(disp1Array)-1] = uf[0]
            disp2Array[len(disp2Array)-1] = uf[1]

        # State determination, starting with Basic displacements
        ub1 = np.dot(Tbf1, uf)
        ub2 = np.dot(Tbf2, uf)
        ub3 = np.dot(Tbf3, uf)

        # Basic forces
        Fb1 = (spring1(ub1))[0]
        Fb2 = (spring2(ub2))[0]
        Fb3 = (spring3(ub3))[0]

        # Final force vector
        tildeFf = np.dot(Tbf1.transpose(), Fb1) \
                  + np.dot(Tbf2.transpose(), Fb2) \
                  + np.dot(Tbf3.transpose(), Fb3)

        # Residual vector and its norm
        Rf = tildeFf - Ff
        resNorm = np.linalg.norm(Rf)

        # Increment counter
        i += 1

# Response plot
plt.ion()
fig, ax = plt.subplots()
ax.plot(disp1Array, loadArray, 'bs-', label='DOF 1')
ax.plot(disp2Array, loadArray, 'ks-', label='DOF 2')
plt.xlabel("Displacement")
plt.ylabel("Load")
plt.title("Load-Displacement Curves")
legend = ax.legend(loc='lower right')
print('\n'"Click somewhere in the plot to continue...")
plt.waitforbuttonpress()
