# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. One objective in this file
# is to demonstrate features and capabilities of the finite element
# software framework OpenSees, here ran via its Python interface,
# OpenSeesPy. There are many good resources online to learn more about
# OpenSees. The Portwood Digital blog published at PortwoodDigital.com
# by Professor Michael Scott at Oregon State is a nice place to start.
# ------------------------------------------------------------------------

# Import external libraries
from sys import platform
if platform == "darwin":
    from openseespymac.opensees import *
elif platform == "win32":
    from openseespy.opensees import *
else:
    print("Cannot handle this type of operating system")
    import sys
    sys.exit()

# Clean up, set 2D structure, set 3 DOFs per node
wipe()
model('basic', '-ndm', 2, '-ndf', 3)

# Input [N, mm]
L = 4000.0
b = 30.0
h = 200.0
E = 200000.0
A = b*h
I = b*h**3/12
P = 4000.0

# Nodes
node(1, 0.0, 0.0)
node(2, L, 0.0)

# Boundary conditions
fix(1, 1, 1, 1)

# Geometric transformation
geomTransf('Linear', 1)

# Element [el n1 n2  A  E  I transf]
element('elasticBeamColumn', 1, 1, 2, A, E, I, 1)

# Create time series in order to apply load
timeSeries('Linear', 1)

# Create load patter in order to apply load [patternTag, timeSeriesTag]
pattern('Plain', 1, 1)

# Nodal load [nd  Fx Fy Mz]
load(2, 0.0, -P, 0.0)

# Analysis setup
system('BandGeneral')            # How to store and solve the system of equations (SOE)
constraints('Plain')             # How boundary conditions are handled
numberer('Plain')                # How DOFs are numbered, e.g., to minimize bandwidth
algorithm('Linear')              # The linear algorithm uses one step to solve the SOE
integrator('LoadControl', 1.0)   # Set load factor increment lambda=1.0
test('NormDispIncr', 1.0e-8, 10) # Constructs the SOE & solver plus checks convergence
analysis('Static')               # Defines analysis type

# Run the analysis
analyze(1)

# Extract the displacement response from OpenSees
uOpenSees = abs(nodeDisp(2, 2))

# Analytical solution
uAnalytical = P * L**3 / (3.0 * E * I)

# Output
print('\n'"OpenSees displacement:   %.2fmm" % uOpenSees)
print("Analytical displacement: %.2fmm" % uAnalytical)
