# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. One objective in this file
# is to demonstrate features and capabilities of the finite element
# software framework OpenSees, here ran via its Python interface,
# OpenSeesPy. There are many good resources online to learn more about
# OpenSees. The Portwood Digital blog published at PortwoodDigital.com
# by Professor Michael Scott at Oregon State is a nice place to start.
# ------------------------------------------------------------------------

# Import external libraries
import numpy as myNumpy
import matplotlib.pyplot as plt
from sys import platform
if platform == "darwin":
    from openseespymac.opensees import *
elif platform == "win32":
    from openseespy.opensees import *
else:
    print("Cannot handle this type of operating system")
    import sys
    sys.exit()

# Clean up, set 2D structure, set 3 DOFs per node
wipe()
model('basic', '-ndm', 2, '-ndf', 3)

# Dimensions
width = 5.0
height = 4.0
colWidth = 0.4
colDepth = 0.6
cover = 0.04
As = 0.000387096  # Area of no. 7 bars in m^2

# Nodes
node(1, 0.0, 0.0)
node(2, width, 0.0)
node(3, 0.0, height)
node(4, width, height)

# Supports
fix(1, 1, 1, 1)
fix(2, 1, 1, 1)

# Core concrete (confined)    tag   f'c      ec0     f'cu    ecu
uniaxialMaterial('Concrete01', 1, -35.0e6, -0.004, -30.0e6, -0.014)

# Cover concrete (unconfined)
uniaxialMaterial('Concrete01', 2, -30.0e6, -0.002, 0.0, -0.006)

# Reinforcing steel        tag   fy      E0           b
uniaxialMaterial('Steel01', 3, 500.0e6, 200000.0e6, 0.01)

# Core fibers
y1 = colDepth / 2.0
z1 = colWidth / 2.0
section('Fiber', 1)
patch('rect', 1, 10, 1, cover - y1, cover - z1, y1 - cover, z1 - cover)

# Cover fibers (top, bottom, left, right)
patch('rect', 2, 10, 1, -y1, z1 - cover, y1, z1)
patch('rect', 2, 10, 1, -y1, -z1, y1, cover - z1)
patch('rect', 2, 2, 1, -y1, cover - z1, cover - y1, z1 - cover)
patch('rect', 2, 2, 1, y1 - cover, cover - z1, y1, z1 - cover)

# Reinforcing fibers (left, middle, right)
layer('straight', 3, 3, As, y1 - cover, z1 - cover, y1 - cover, cover - z1)
layer('straight', 3, 2, As, 0.0, z1 - cover, 0.0, cover - z1)
layer('straight', 3, 3, As, cover - y1, z1 - cover, cover - y1, cover - z1)

# Lobatto integration
np = 5
beamIntegration('Lobatto', 1, 1, np)

# Include P-delta effects in columns
geomTransf('PDelta', 1)

# Columns   tag ndI ndJ transfTag integrationTag
element('forceBeamColumn', 1, 1, 3, 1, 1)
element('forceBeamColumn', 2, 2, 4, 1, 1)

# Skip P-delta effects in the beam
geomTransf('Linear', 2)

# Beam                     tag, ndI, ndJ, A, E, Iz, transfTag
element('elasticBeamColumn', 3, 3, 4, 360.0, 4030.0, 8640.0, 2)

# Gravity loads, roughly 10% of the axial capacity of each column
P = 800.0e3

# Create a plain load pattern with a linear time series
timeSeries('Linear', 1)
pattern('Plain', 1, 1)

# Create nodal loads at nodes 3 & 4
#   nd  FX,  FY, MZ
load(3, 0.0, -P, 0.0)
load(4, 0.0, -P, 0.0)

# ------------------------------------------------------------------------
# OPENSEES ANALYSIS SETUP FOR GRAVITY LOAD
# ------------------------------------------------------------------------

# Create the system of equation, a sparse solver with partial pivoting
system('BandGeneral')

# Create the constraint handler, the transformation method
constraints('Transformation')

# Create the DOF numberer, the reverse Cuthill-McKee algorithm
numberer('RCM')

# Create the convergence test, the norm of the residual with a tolerance of
# 1e-12 and a max number of iterations of 10
test('NormDispIncr', 1.0e-12, 10, 3)

# Create the solution algorithm, a Newton-Raphson algorithm
algorithm('Newton')

# Create the integration scheme, the LoadControl scheme using steps of 0.1
integrator('LoadControl', 0.1)

# Create the analysis object
analysis('Static')

# ------------------------------------------------------------------------
# RUN OPENSEES FOR GRAVITY LOADS
# ------------------------------------------------------------------------

# Ten load steps
analyze(10)

# This is how to obtain nodal displacements
u3 = nodeDisp(3, 2)
u4 = nodeDisp(4, 2)

# ------------------------------------------------------------------------
# PREPARE OPENSEES FOR GROUND MOTION
# ------------------------------------------------------------------------

# Set the gravity loads to be constant & reset the time in the domain
loadConst('-time', 0.0)

# Define nodal mass in terms of axial load on columns
g = 9.81
m = P / g
mass(3, m, m, 0.0)
mass(4, m, m, 0.0)

# Set time series to be passed to uniform excitation
timeSeries('Path', 2, '-filePath', 'GroundMotionElCentro.txt', '-dt', 0.02, '-factor', g)

# UniformExcitation load pattern
#                          tag dir
pattern('UniformExcitation', 2, 1, '-accel', 2)

# Eigenvalue analysis before shaking
numEigen = 2
eigenValues = eigen(numEigen)
omega1 = myNumpy.sqrt(eigenValues[0])
omega2 = myNumpy.sqrt(eigenValues[1])
print('\n'"First two natural periods BEFORE shaking:", 2*myNumpy.pi/omega1, 2*myNumpy.pi/omega2)

# Rayleigh damping (5% damping on first two modes, using initial stiffness)
factor = 2 * 0.05 / (omega1 + omega2)
cM = omega1 * omega2 * factor
cK = factor
rayleigh(cM, 0.0, cK, 0.0)

# Delete the old analysis and all it's component objects
wipeAnalysis()

# ------------------------------------------------------------------------
# OPENSEES ANALYSIS SETUP FOR DYNAMIC ANALYSIS
# ------------------------------------------------------------------------

# Create the system of equation, a banded general storage scheme
system('BandGeneral')

# Create the constraint handler, a plain handler as homogeneous boundary
constraints('Plain')

# Create the convergence test, the norm of the residual with a tolerance of
# 1e-12 and a max number of iterations of 10
test('NormDispIncr', 1.0e-12, 10)

# Create the solution algorithm, a Newton-Raphson algorithm
algorithm('Newton')

# Create the DOF numberer, the reverse Cuthill-McKee algorithm
numberer('RCM')

# Create the integration scheme, the Newmark with alpha =0.5 and beta =.25
integrator('Newmark', 0.5, 0.25)

# Create the analysis object
analysis('Transient')


# ------------------------------------------------------------------------
# RUN OPENSEES DYNAMIC ANALYSIS
# ------------------------------------------------------------------------

tCurrent = getTime()
ok = 0

time = [tCurrent]
u3 = [0.0]

duration = 35
while ok == 0 and tCurrent < duration:

    ok = analyze(1, .01)

    # if the analysis fails try initial tangent iteration
    if ok != 0:
        print("regular newton failed .. lets try an initail stiffness for this step")
        test('NormDispIncr', 1.0e-12, 100, 0)
        algorithm('ModifiedNewton', '-initial')
        ok = analyze(1, .01)
        if ok == 0:
            print("that worked .. back to regular newton")
        test('NormDispIncr', 1.0e-12, 10)
        algorithm('Newton')

    tCurrent = getTime()

    time.append(tCurrent)
    u3.append(nodeDisp(3, 1))


# Eigenvalue analysis after shaking
eigenValues = eigen(numEigen)
omega1 = myNumpy.sqrt(eigenValues[0])
omega2 = myNumpy.sqrt(eigenValues[1])
print('\n'"First two natural periods AFTER shaking:", 2*myNumpy.pi/omega1, 2*myNumpy.pi/omega2)


if ok == 0:
    print('\n'"OpenSees dynamic analysis PASSED")
else:
    print('\n'"OpenSees dynamic analysis FAILED")

# ------------------------------------------------------------------------
# PLOT THE RESPONSE
# ------------------------------------------------------------------------

plt.ion()
plt.figure()
plt.plot(time, u3, 'k-')
plt.xlabel("Time [seconds]")
plt.title("Horizontal Displacement of Node 3 [metres]")
print('\n'"Click somewhere in the plot to continue...")
plt.waitforbuttonpress()
