# ------------------------------------------------------------------------
# The following Python code is implemented by Professor Terje Haukaas at
# the University of British Columbia in Vancouver, Canada. It is made
# freely available online at terje.civil.ubc.ca together with notes,
# examples, and additional Python code. One objective in this file
# is to demonstrate features and capabilities of the finite element
# software framework OpenSees, here ran via its Python interface,
# OpenSeesPy. There are many good resources online to learn more about
# OpenSees. The Portwood Digital blog published at PortwoodDigital.com
# by Professor Michael Scott at Oregon State is a nice place to start.
# ------------------------------------------------------------------------

# Import external libraries
import numpy as np
import matplotlib.pyplot as plt
from sys import platform
if platform == "darwin":
    from openseespymac.opensees import *
elif platform == "win32":
    from openseespy.opensees import *
else:
    print("Cannot handle this type of operating system")
    import sys
    sys.exit()

# Clean up, set 2D structure, set 3 DOFs per node
wipe()
model('basic', '-ndm', 2, '-ndf', 3)

# Analysis parameters
numIncrements = 20
dt = 1.0 / numIncrements

# Steel01 material [N, mm]
E = 210000.0         # Modulus of elasticity
fy = 350.0           # Yield stress
alpha = 0.03         # Second-slope stiffness is Ep=E*alpha
uniaxialMaterial('Steel01', 1, fy, E, alpha)

# Wide-flange cross-section section('WFSection2d', secTag, matTag, d, tw, bf, tf, Nfw, Nff)
hw = 250.0           # Web height
bf = 250.0           # Flange width
tf = 15.0            # Flange thickness
tw = 15.0            # Web thickness
nf = 3               # Number of fibers in the flange
nw = 8               # Number of fibres in the web
A = tw*(hw-2*tf) + 2*bf*tf
I = tw*(hw-2*tf)**3/12.0 + 2*bf*tf*(0.5*(hw-tf))**2
section('WFSection2d', 1, 1, hw, tw, bf, tf, nw, nf)

# Nodes
geometricalImperfection = 40.0
geometricalImperfection = 0.0
nel = 5              # Number of elements along cantilever
L = 3500.0           # Total length of cantilever
for i in range(nel+1):
    node(i+1, i*geometricalImperfection/nel, i*L/nel)

#  Fixed base
fix(1, 1, 1, 1)

# Integration rule beamIntegration('Lobatto', tag, secTag, N)
beamIntegration('Lobatto', 1, 1, 5)

# Dealing with the axial force
P = 0.8 * (np.pi**2 * E * I / (2.0*L)**2)
P = 0
print('\n'"Axial crushing capacity: %.1fkN" % (A*fy/1000.0))
print("Buckling load:           %.1fkN" % (np.pi**2 * E * I / (2.0*L)**2 / 1000.0))
print("Applied axial load:      %.1fkN" % (P/1000.0))
geomTransf('PDelta', 1)  # Linear / PDelta
timeSeries("Linear", 1)  # Constant / Linear
pattern("Plain", 1, 1)
load(nel+1, 0.0, -P, 0.0)

# Create elements element('dispBeamColumn', eleTag, *eleNodes, transfTag, integrationTag, '-cMass', '-mass', mass=0.0)
for i in range(nel):
    element('dispBeamColumn', i+1, i+1, i+2, 1, 1)
    #element('elasticBeamColumn', i+1, i+1, i+2, A, E, I, 1)

# Lateral load
F = 0.0
F = ((2*fy*I/(hw*L))) * 2
print('\n'"Lateral yield load:   %.1fkN" % ((2*fy*I/(hw*L))/1000.0))
print("Applied lateral load: %.1fkN" % (F/1000.0))
timeSeries("Linear", 2)
pattern("Plain", 2, 2)
load(nel+1, F, 0.0, 0.0)

# Analysis setup
system("ProfileSPD")
numberer("Plain")
constraints("Plain")
integrator('LoadControl', dt)
algorithm("Newton")
test('NormDispIncr', 1e-8, 100, 0)
analysis("Static")

# Perform analysis
timeArray = [0.0]
dispArray = [0.0]
for i in range(numIncrements):

    # Try one increment
    ok = analyze(1)

    # Switch to displacement control if the increment didn't converge
    if ok != 0:

        # Inform the user
        print('\n'"Failed to converge at time %.2f. Switching to displacement control..." % getTime())

        # Switch to displacement control
        integrator('DisplacementControl', nel+1, 1, 10.0) # Displacement increment 10mm at node nel+1 DOF 1

        # Try one increment now
        ok = analyze(1)
        if ok != 0:
            print("Still could not converge...")

    # Store response for plotting
    timeArray.append(getTime())
    dispArray.append(nodeDisp(nel+1, 1))

# Print top displacement response
print('\n'"OpenSees top displacement:   %.1fmm" % nodeDisp(nel+1, 1))
print("Linear elastic displacement: %.1fmm" % (F*L**3/(3*E*I)))

# Plot response
plt.ion()
plt.plot(dispArray, timeArray, 'bs-', label='Top displacement')
plt.xlabel("Displacement")
plt.ylabel("Pseudo Time")
plt.title("Response")
print('\n'"Click somewhere in the plot to continue...")
plt.waitforbuttonpress()
